# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, cast
from collections.abc import Iterable
from types import MappingProxyType
from slyme.utils.pytree import (
    AttributeKey,
    PyTreeEngine,
    PyTreeAux,
    MappingKey,
    PYTREE_ENGINE_REGISTRY,
)
from slyme.utils.pytree.common import flatten_mapping_proxy, unflatten_mapping_proxy
from .core import ContextData, Context

# Context engine
CONTEXT_ENGINE = PyTreeEngine("context_engine", register_defaults=False)
PYTREE_ENGINE_REGISTRY.register(CONTEXT_ENGINE, key="context_engine")


def _flatten_context_data(data: ContextData) -> tuple[Iterable[Any], PyTreeAux]:
    """Flatten ContextData."""
    keys = tuple(data.keys())
    rich_keys = tuple(MappingKey(k) for k in keys)
    children = (data[k] for k in keys)
    return children, PyTreeAux(children_keys=rich_keys)


def _unflatten_context_data(children: Iterable[Any], aux: PyTreeAux) -> ContextData:
    """Unflatten to ContextData."""
    if aux.children_keys is None:
        raise ValueError("Missing keys for ContextData unflattening.")
    raw_keys = [k.key for k in cast("Iterable[MappingKey]", aux.children_keys)]
    return ContextData(zip(raw_keys, children))


def _flatten_context(context: Context) -> tuple[Iterable[Any], PyTreeAux]:
    """Flatten Context -> (root_context_data, )."""
    return (context._root,), PyTreeAux(
        metadata={"hook": context._hook}, children_keys=(AttributeKey("_root"),)
    )


def _unflatten_context(children: Iterable[Any], aux: PyTreeAux) -> Context:
    """Unflatten Context."""
    (root,) = children
    return Context._from_context_data(root, hook=aux.metadata["hook"])


CONTEXT_ENGINE.register(ContextData, _flatten_context_data, _unflatten_context_data)
CONTEXT_ENGINE.register(Context, _flatten_context, _unflatten_context)
# Context eval
CTX_EVAL_ENGINE = PyTreeEngine("ctx_eval_engine", register_defaults=True)
PYTREE_ENGINE_REGISTRY.register(CTX_EVAL_ENGINE, key="ctx_eval_engine")
CTX_EVAL_ENGINE.register(
    MappingProxyType, flatten_mapping_proxy, unflatten_mapping_proxy
)



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Protocol, runtime_checkable


@runtime_checkable
class HasExtraRepr(Protocol):
    """
    Protocol for objects that provide an extended string representation.

    This allows objects to expose internal state details for logging,
    debugging, or rendering without overriding the standard __repr__.
    """

    def extra_repr(self) -> str:
        """Return the extra representation string."""
        ...


@runtime_checkable
class HasTypeRepr(Protocol):
    """
    Protocol for objects that provide a custom type name for rendering.

    This allows objects (like functional wrappers) to masquerade as their
    underlying logic (e.g. function name) instead of their implementation class.
    """

    def type_repr(self) -> str:
        """Return the custom type name string."""
        ...



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps, partial
from enum import Enum
from typing import (
    TypeVar,
    Callable,
    Union,
    overload,
)
from typing_extensions import ParamSpec
from slyme.node import Node, check_node_structure

_NodeT = TypeVar("_NodeT", bound=Node)
_P = ParamSpec("_P")
# Marker for missing arguments to handle @builder vs @builder()
_Missing = Enum("_Missing", ["MARK"])
_MISSING = _Missing.MARK


def _builder(
    func: Callable[_P, _NodeT],
    /,
    *,
    check_structure: bool = True,
) -> Callable[_P, _NodeT]:
    @wraps(func)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _NodeT:
        node = func(*args, **kwargs)
        if node is None:
            raise ValueError(
                f"The builder function '{func.__name__}' returned None. "
                "Did you forget to return the constructed Node?"
            )

        if check_structure:
            check_node_structure(node)
        return node

    return wrapper


@overload
def builder(
    func: _Missing = _MISSING,
    /,
    *,
    check_structure: bool = True,
) -> Callable[[Callable[_P, _NodeT]], Callable[_P, _NodeT]]: ...
@overload
def builder(
    func: Callable[_P, _NodeT],
    /,
    *,
    check_structure: bool = True,
) -> Callable[_P, _NodeT]: ...
def builder(
    func: Union[Callable[_P, _NodeT], _Missing] = _MISSING,
    /,
    *,
    check_structure: bool = True,
) -> Union[
    Callable[_P, _NodeT],
    Callable[[Callable[_P, _NodeT]], Callable[_P, _NodeT]],
]:
    """
    Decorator to create a builder function.
    It wraps the function to ensure it returns a valid Node and optionally checks the structure.
    """
    if func is _MISSING:
        return partial(_builder, check_structure=check_structure)
    else:
        return _builder(func, check_structure=check_structure)



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .core import Context, Ref

__all__ = [
    "ExtractResult",
    "MutateResult",
    "Hook",
    "HookChain",
]


@dataclass(frozen=True)
class ExtractResult:
    values: tuple[Any, ...]


@dataclass(frozen=True)
class MutateResult:
    updates: dict["Ref", Any]
    drops: set["Ref"]


@dataclass(frozen=True)
class Hook:
    def on_extract(
        self,
        *,
        ctx: "Context",
        refs: tuple["Ref", ...],
        values: tuple[Any, ...],
        **kwargs,
    ) -> ExtractResult:
        return ExtractResult(values=values)

    async def on_async_extract(
        self,
        *,
        ctx: "Context",
        refs: tuple["Ref", ...],
        values: tuple[Any, ...],
        **kwargs,
    ) -> ExtractResult:
        return self.on_extract(ctx=ctx, refs=refs, values=values, **kwargs)

    def on_mutate(
        self, *, ctx: "Context", updates: dict["Ref", Any], drops: set["Ref"], **kwargs
    ) -> MutateResult:
        return MutateResult(updates=updates, drops=drops)

    async def on_async_mutate(
        self, *, ctx: "Context", updates: dict["Ref", Any], drops: set["Ref"], **kwargs
    ) -> MutateResult:
        return self.on_mutate(ctx=ctx, updates=updates, drops=drops, **kwargs)


@dataclass(frozen=True)
class HookChain(Hook):
    hooks: tuple[Hook, ...]

    def on_extract(
        self,
        *,
        ctx: "Context",
        refs: tuple["Ref", ...],
        values: tuple[Any, ...],
        **kwargs,
    ) -> ExtractResult:
        for hook in self.hooks:
            result = hook.on_extract(ctx=ctx, refs=refs, values=values, **kwargs)
            values = result.values
        return ExtractResult(values=values)

    async def on_async_extract(
        self,
        *,
        ctx: "Context",
        refs: tuple["Ref", ...],
        values: tuple[Any, ...],
        **kwargs,
    ) -> ExtractResult:
        for hook in self.hooks:
            result = await hook.on_async_extract(
                ctx=ctx, refs=refs, values=values, **kwargs
            )
            values = result.values
        return ExtractResult(values=values)

    def on_mutate(
        self, *, ctx: "Context", updates: dict["Ref", Any], drops: set["Ref"], **kwargs
    ) -> MutateResult:
        for hook in self.hooks:
            result = hook.on_mutate(ctx=ctx, updates=updates, drops=drops, **kwargs)
            updates, drops = result.updates, result.drops
        return MutateResult(updates=updates, drops=drops)

    async def on_async_mutate(
        self, *, ctx: "Context", updates: dict["Ref", Any], drops: set["Ref"], **kwargs
    ) -> MutateResult:
        for hook in self.hooks:
            result = await hook.on_async_mutate(
                ctx=ctx, updates=updates, drops=drops, **kwargs
            )
            updates, drops = result.updates, result.drops
        return MutateResult(updates=updates, drops=drops)



from slyme.node import node, Auto, sequential, expression
from slyme.context import Context, Ref


@node
def node1(
    ctx: Context,
    /,
    *,
    value: Ref[str],
):
    return ctx.set(value, "hello world!")


@node
def node2(
    ctx: Context,
    /,
    *,
    value: Auto[str] = "i'm here",
):
    print(value)
    return ctx


@expression
def expression(
    ctx: Context,
    /,
    *,
    value: Auto[str],
):
    return value * 3


scope = {
    "value": Ref("hello.value")
}


root = sequential(nodes=[
    node1(scope),
    node2(scope, value=expression(scope)),
])
breakpoint()
root_exec = root.prepare()
breakpoint()
root_exec(Context())
import transformers



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from abc import ABC, abstractmethod
from enum import Enum
from dataclasses import dataclass, field, InitVar
from collections import defaultdict
from collections.abc import Iterable, Mapping, Callable
from typing import (
    Any,
    Generic,
    TypeVar,
    Union,
    Literal,
    cast,
    Optional,
    overload,
    TYPE_CHECKING,
)
from typing_extensions import Self
from slyme.utils.pytree import (
    PyTreeEngine,
    MappingKey,
    KeyPath,
)
from slyme.utils.exception import enrich_exception

if TYPE_CHECKING:
    from .hook import Hook

_T = TypeVar("_T")
_T2 = TypeVar("_T2")
_EMPTY_MAPPING = types.MappingProxyType({})
_Missing = Enum("_Missing", ["MARK"])
_MISSING = _Missing.MARK
DiffMissing = Enum("DiffMissing", ["MARK"])
DIFF_MISSING = DiffMissing.MARK


class Config:
    repr_indent: str
    repr_newline: str
    repr_suffix: str
    repr_last_suffix: str
    leaf_formatter: Callable[[Any], str]

    @classmethod
    def set_compact_repr(cls) -> type[Self]:
        cls.repr_indent = ""
        cls.repr_newline = ""
        cls.repr_suffix = ", "
        cls.repr_last_suffix = ""
        return cls

    @classmethod
    def set_pretty_repr(cls) -> type[Self]:
        cls.repr_indent = "    "
        cls.repr_newline = "\n"
        cls.repr_suffix = ","
        cls.repr_last_suffix = ","
        return cls

    @classmethod
    def set_truncated_repr(cls, max_len: int = 100) -> type[Self]:
        def _truncated(obj: Any) -> str:
            s = repr(obj)
            return s if len(s) <= max_len else s[:max_len] + "..."

        cls.leaf_formatter = _truncated
        return cls


Config.set_pretty_repr().set_truncated_repr(max_len=100)


@dataclass(frozen=True, repr=False, eq=False)
class Ref(Generic[_T]):
    """Immutable dotted ref with cached hash and split parts.

    NOTE: Ref.key_path can only resolve leaf values.
    """

    path: str
    key_path: KeyPath = ()
    metadata: Mapping[str, Any] = field(default_factory=lambda: _EMPTY_MAPPING)
    parts: tuple[str, ...] = field(init=False)
    hash: int = field(init=False)

    def __post_init__(self) -> None:
        if not self.path:
            raise ValueError("Empty ref path")
        parts = tuple(self.path.split("."))
        if any(not p for p in parts):
            raise ValueError(f"Invalid ref path: {self.path!r}")
        # Bypass frozen=True to set fields
        object.__setattr__(self, "parts", parts)
        object.__setattr__(self, "hash", hash((parts, self.key_path)))
        if not isinstance(self.metadata, types.MappingProxyType):
            object.__setattr__(self, "metadata", types.MappingProxyType(self.metadata))

    def _resolve(self, pytree) -> _T:
        return PyTreeEngine.get_element(pytree, self.key_path)

    def update_metadata(self, metadata: Mapping[str, Any]) -> "Ref[_T]":
        """Returns a new Ref with updated metadata (merging with existing)."""
        new_metadata = dict(self.metadata)
        new_metadata.update(metadata)
        return Ref(self.path, key_path=self.key_path, metadata=new_metadata)

    def at(
        self,
        subpath: str,
        key_path: Union[KeyPath, _Missing] = _MISSING,
        metadata: Union[Optional[Mapping[str, Any]], _Missing] = _MISSING,
    ) -> "Ref":
        """
        Create a new Ref at a subpath relative to this Ref.

        Does NOT inherit key_path or metadata from the parent Ref by default.
        """
        new_path = f"{self.path}.{subpath}" if self.path else subpath
        kwargs = {}
        if key_path is not _MISSING:
            kwargs["key_path"] = key_path
        if metadata is not _MISSING:
            kwargs["metadata"] = metadata
        return Ref(new_path, **kwargs)

    def __hash__(self) -> int:
        return self.hash

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Ref)
            and self.parts == other.parts
            and self.key_path == other.key_path
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.extra_repr()})"

    def extra_repr(self) -> str:
        repr_items = [f"path={self.path!r}"]
        if self.key_path:
            repr_items.append(
                f"key_path_expr={PyTreeEngine.codify_key_path(self.key_path)}"
            )
        if self.metadata:
            repr_items.append(f"metadata={self.metadata!r}")
        return ", ".join(repr_items)


class ContextPathError(KeyError):
    """Internal exception raised when a path cannot be resolved in the context."""

    pass


class ContextData(dict[str, Any]):
    """
    Internal dictionary implementation used to distinguish structural elements
    from user-provided dictionary values.

    Acts as the core "Smart Node" for the Context, handling recursive Copy-On-Write logic.
    """

    __slots__ = ()

    def _raise_immutable(self, *args, **kwargs):
        raise TypeError(
            f"{type(self).__name__} is immutable. "
            "Use functional modifications (e.g. context.mutate/update/drop) to create a new instance."
        )

    # Disable all mutable methods via assignment to minimize boilerplate
    __setitem__ = _raise_immutable
    __delitem__ = _raise_immutable
    pop = _raise_immutable
    popitem = _raise_immutable
    clear = _raise_immutable
    update = _raise_immutable
    setdefault = _raise_immutable
    # Disable in-place operators
    __ior__ = _raise_immutable

    def mutate(
        self,
        updates: dict[tuple[str, ...], Any],
        drops: set[tuple[str, ...]],
    ) -> "ContextData":
        """
        Core recursive Copy-On-Write (COW) algorithm for simultaneous updates and drops.

        Semantics:
        - Logically, drops are executed first, then updates.
          However, for efficiency, they are processed in a single pass.
        - Updates have higher priority than drops if they target the same path or its children.
          (i.e., updating 'a' overrides dropping 'a' or dropping 'a.b').
        - Drop Conflicts:
          - Overlapping drops (e.g., drop 'a' and drop 'a.b') are handled implicitly.
            Dropping the parent 'a' removes the entire subtree, so dropping 'a.b' is redundant but valid.
        - Update Conflicts:
          - Overlapping updates (e.g., update 'a' and update 'a.b') raise a ValueError.
            We cannot determine whether 'a' should be a leaf (value) or a container (for 'b').

        Algorithm:
        1. Base Cases:
           - If root is updated, return new value immediately.
           - If root is dropped (and not updated), return _MISSING.
        2. Group Operations:
           - Group updates and drops by their head key (first path component).
        3. Recursive Application:
           - For each head, recursively call mutate on the child.
           - Handle leaf conflicts (path blocked by existing value).
           - Enhance exceptions with path context.
        """
        # 1. Base Cases
        if () in updates:
            return updates[()]

        # Determine Base State (Reset vs Copy)
        if () in drops:
            # Reset: Start from empty. Existing data is discarded.
            if not updates:
                return _MISSING
            new_data = {}
        else:
            # Copy: Start from existing.
            if not updates and not drops:
                return self
            new_data = dict(self)

        # 2. Group Operations
        grouped_ops = defaultdict(lambda: ({}, set()))

        for path, val in updates.items():
            if not path:
                continue  # Handled above
            head, *tail = path
            grouped_ops[head][0][tuple(tail)] = val

        for path in drops:
            if not path:
                continue  # Handled above
            head, *tail = path
            grouped_ops[head][1].add(tuple(tail))

        # 3. Recursive Application
        for head, (sub_updates, sub_drops) in grouped_ops.items():
            # Optimization: Exact overwrite
            if () in sub_updates:
                # Conflict: Cannot update both parent and child simultaneously
                if len(sub_updates) > 1:
                    raise ValueError(
                        f"Update conflict at '{head}': Cannot update both parent and child simultaneously."
                    )

                # Apply update
                new_data[head] = sub_updates[()]

                # If we are overwriting the head, any drops for this head or its children
                # are implicitly handled (superseded by the overwrite).
                # This satisfies "batch drops then batch updates" semantics.
                continue

            # Get existing child or MISSING (if Reset, it's always MISSING)
            child = new_data.get(head, _MISSING)

            # Structure Validation / Auto-Vivification
            if isinstance(child, ContextData):
                pass
            elif child is _MISSING:
                # Create new container for updates
                if not sub_updates:
                    continue
                child = ContextData()
            elif () in sub_drops:
                # Child is a leaf.
                # If we are dropping the leaf itself, allow it.
                if not sub_updates:
                    new_data.pop(head, None)
                    continue
                # If we also have updates, we start from a fresh container (Reset)
                child = ContextData()
            else:
                # Conflict: Path blocked by leaf value
                raise ContextPathError(
                    f"Path '{head}' blocked by leaf value during mutation."
                )

            # Enrich both ContextPathError (structural issues) and ValueError (update conflicts)
            with enrich_exception(head, exc_types=(ContextPathError, ValueError)):
                new_child = child.mutate(sub_updates, sub_drops)
                if new_child is _MISSING:
                    new_data.pop(head, None)
                else:
                    new_data[head] = new_child

        return ContextData(new_data)


@dataclass(frozen=True)
class ContextDiff:
    """
    Recursive diff structure representing changes between two ContextElements.
    Supports nested diffs for containers and direct value changes for leaves.
    """

    added: Mapping[str, Any] = field(default_factory=lambda: _EMPTY_MAPPING)
    removed: Mapping[str, Any] = field(default_factory=lambda: _EMPTY_MAPPING)
    modified: Mapping[str, tuple[Any, Any]] = field(
        default_factory=lambda: _EMPTY_MAPPING
    )
    nested: Mapping[str, "ContextDiff"] = field(default_factory=lambda: _EMPTY_MAPPING)

    def __post_init__(self):
        for name in ("added", "removed", "modified", "nested"):
            val = getattr(self, name)
            if not isinstance(val, types.MappingProxyType):
                object.__setattr__(self, name, types.MappingProxyType(val))

    def __bool__(self) -> bool:
        return bool(self.added or self.removed or self.modified or self.nested)

    def __repr__(self) -> str:
        parts = []
        if self.added:
            parts.append(f"added={list(self.added.keys())}")
        if self.removed:
            parts.append(f"removed={list(self.removed.keys())}")
        if self.modified:
            parts.append(f"modified={list(self.modified.keys())}")
        if self.nested:
            parts.append(f"nested={list(self.nested.keys())}")
        if not parts:
            return "ContextDiff(no changes)"
        return f"ContextDiff({', '.join(parts)})"

    def flatten(self) -> dict[str, tuple[Any, Any]]:
        """
        Flatten the recursive diff into a single dictionary of changes.
        Returns a dict of {path: (old_value, new_value)}.
        Added values have old_value as DIFF_MISSING.
        Removed values have new_value as DIFF_MISSING.
        """
        changes: dict[str, tuple[Any, Any]] = {}

        for k, v in self.added.items():
            changes[k] = (DIFF_MISSING, v)
        for k, v in self.removed.items():
            changes[k] = (v, DIFF_MISSING)
        for k, v in self.modified.items():
            changes[k] = v

        for key, child_diff in self.nested.items():
            for child_path, (old, new) in child_diff.flatten().items():
                changes[f"{key}.{child_path}"] = (old, new)

        return changes


def diff_context_data(
    left: Mapping[str, Any],
    right: Mapping[str, Any],
    strategy: Literal["is", "eq"] = "is",
) -> ContextDiff:
    """
    Compute the recursive difference between two ContextData objects.
    Optimized for Copy-On-Write (COW) structures by skipping identical subtrees.
    """
    if left is right:
        return ContextDiff()

    added = {}
    removed = {}
    modified = {}
    nested = {}

    keys_left = set(left.keys())
    keys_right = set(right.keys())

    # Removed keys
    for k in keys_left - keys_right:
        removed[k] = left[k]

    # Added keys
    for k in keys_right - keys_left:
        added[k] = right[k]

    # Shared keys
    for k in keys_left & keys_right:
        val_left = left[k]
        val_right = right[k]

        # Optimization: Skip identical objects (COW sharing)
        if val_left is val_right:
            continue

        # Check for nested containers (ContextData)
        is_container_left = isinstance(val_left, ContextData)
        is_container_right = isinstance(val_right, ContextData)

        if is_container_left and is_container_right:
            child_diff = diff_context_data(val_left, val_right, strategy)
            if child_diff:
                nested[k] = child_diff
        else:
            # Leaf comparison
            is_different = False
            if strategy == "is":
                is_different = val_left is not val_right
            elif strategy == "eq":
                is_different = val_left != val_right

            if is_different:
                modified[k] = (val_left, val_right)

    return ContextDiff(added, removed, modified, nested)


class ContextElement(ABC):
    """
    Abstract base class for context-related entities (Context, ContextView).
    """

    __slots__ = ()

    @abstractmethod
    def extract(self, ref_tree: Any, *, apply_hook: bool = True, **kwargs) -> Any:
        pass

    @abstractmethod
    async def async_extract(
        self, ref_tree: Any, *, apply_hook: bool = True, **kwargs
    ) -> Any:
        pass

    @abstractmethod
    def get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        pass

    @abstractmethod
    async def async_get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        pass

    @abstractmethod
    def exists(self, ref: Ref[_T]) -> bool:
        pass

    @abstractmethod
    def keys(self, ref: Optional[Ref[_T]] = None) -> Iterable[str]:
        pass

    @abstractmethod
    def to_context_data(self, ref: Optional[Ref[_T]] = None) -> ContextData:
        pass

    @abstractmethod
    def to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        """Convert to standard python dictionary recursively."""
        pass

    @abstractmethod
    async def async_to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        pass

    def type_repr(self) -> str:
        return type(self).__name__

    def __repr__(self) -> str:
        name = self.type_repr()
        try:
            keys = list(self.keys())
        except ContextPathError:
            return f"{name}(<invalid path>)"

        if not keys:
            return f"{name}()"

        newline = Config.repr_newline
        indent = Config.repr_indent
        formatter = Config.leaf_formatter

        item_blocks = []
        for key in keys:
            val = self.get(Ref(key), apply_hook=False)
            v_str = repr(val) if isinstance(val, ContextElement) else formatter(val)

            if newline:
                v_lines = v_str.split(newline)
                if len(v_lines) > 1 and not v_lines[-1]:
                    v_lines.pop()
            else:
                v_lines = [v_str]

            block = [f"{indent}{key!r}: {v_lines[0]}"]
            block.extend(f"{indent}{line}" for line in v_lines[1:])
            item_blocks.append(block)

        body_lines = []
        count = len(item_blocks)
        for i, block in enumerate(item_blocks):
            suffix = Config.repr_last_suffix if i == count - 1 else Config.repr_suffix
            block[-1] += suffix
            body_lines.extend(block)

        return f"{name}({{{newline}{newline.join(body_lines)}{newline}}})"

    def diff(
        self, other: "ContextElement", strategy: Literal["is", "eq"] = "is"
    ) -> ContextDiff:
        """
        Compute the difference between this element and another.
        Returns a recursive ContextDiff structure.
        """
        if strategy not in ("is", "eq"):
            raise ValueError(f"Unknown diff strategy: {strategy!r}")

        # Convert both to ContextData for recursive comparison
        # This handles Context, ContextView, and any other ContextElement
        data_self = self.to_context_data()
        data_other = other.to_context_data()

        return diff_context_data(data_self, data_other, strategy)

    def collect_leaves(self) -> dict[str, Any]:
        # CONTEXT_PYTREE_ENGINE is configured to only traverse Context/ContextData structure.
        # So it treats user dicts as leaves automatically.
        # Use iter_with_key_path for memory efficiency (generator based)
        iterator = CONTEXT_ENGINE.iter_with_key_path(self.to_context_data())

        leaves = {}
        for key_path, leaf in iterator:
            # We know ContextData keys are MappingKeys wrapping strings
            parts = [str(cast(MappingKey, k).key) for k in key_path]
            leaves[".".join(parts)] = leaf
        return leaves


@dataclass(frozen=True, repr=False)
class Context(ContextElement):
    """
    Immutable Context implementation with efficient Copy-On-Write (COW) updates.
    Wraps a root `ContextData`.

    The main **runtime** context container for the node execution.
    """

    _root: ContextData = field(init=False)
    _hook: Optional["Hook"] = field(default=None, init=False)
    data: InitVar[Optional[Mapping[str, Any]]] = None
    hook: InitVar[Optional["Hook"]] = None

    def __post_init__(
        self, data: Optional[Mapping[str, Any]] = None, hook: Optional["Hook"] = None
    ) -> None:
        if data is None:
            root = ContextData()
        elif isinstance(data, ContextData):
            root = data
        else:
            # Shallow conversion strictly for the top level.
            # Trusts user input for deep structure.
            root = ContextData(data)
        object.__setattr__(self, "_root", root)
        object.__setattr__(self, "_hook", hook)

    @classmethod
    def _from_context_data(
        cls, root: ContextData, hook: Optional["Hook"] = None
    ) -> "Context":
        obj = object.__new__(cls)
        object.__setattr__(obj, "_root", root)
        object.__setattr__(obj, "_hook", hook)
        return obj

    def extract(self, ref_tree: Any, *, apply_hook: bool = True, **kwargs) -> Any:
        refs, treedef = CTX_EVAL_ENGINE.flatten(ref_tree)
        values = tuple(self._resolve(ref.parts) for ref in refs)
        # NOTE: Check key_path
        for ref, val in zip(refs, values):
            if isinstance(val, ContextData) and ref.key_path:
                raise ValueError("Ref.key_path can only resolve leaf values.")

        if apply_hook and self._hook:
            result = self._hook.on_extract(ctx=self, refs=refs, values=values, **kwargs)
            values = result.values

        values = tuple(
            ContextView(self, ref.parts)
            if isinstance(val, ContextData)
            else ref._resolve(val)
            for ref, val in zip(refs, values)
        )
        return CTX_EVAL_ENGINE.unflatten(treedef, values)

    async def async_extract(
        self, ref_tree: Any, *, apply_hook: bool = True, **kwargs
    ) -> Any:
        refs, treedef = CTX_EVAL_ENGINE.flatten(ref_tree)
        values = tuple(self._resolve(ref.parts) for ref in refs)
        # NOTE: Check key_path
        for ref, val in zip(refs, values):
            if isinstance(val, ContextData) and ref.key_path:
                raise ValueError("Ref.key_path can only resolve leaf values.")

        if apply_hook and self._hook:
            result = await self._hook.on_async_extract(
                ctx=self, refs=refs, values=values, **kwargs
            )
            values = result.values

        values = tuple(
            ContextView(self, ref.parts)
            if isinstance(val, ContextData)
            else ref._resolve(val)
            for ref, val in zip(refs, values)
        )
        return CTX_EVAL_ENGINE.unflatten(treedef, values)

    def _build_dict_ref_tree(self, ref: Optional[Ref[_T]] = None) -> Any:
        data = self.to_context_data(ref)
        base_path = ref.path if ref else ""

        def build_tree(current_data: ContextData, current_path: str) -> dict[str, Any]:
            tree = {}
            for k, v in current_data.items():
                path = f"{current_path}.{k}" if current_path else k
                if isinstance(v, ContextData):
                    tree[k] = build_tree(v, path)
                else:
                    tree[k] = Ref(path)
            return tree

        return build_tree(data, base_path)

    # --- Read Operations ---
    @overload
    def get(self, ref: Ref[_T], *, apply_hook: bool = True) -> _T: ...
    @overload
    def get(
        self, ref: Ref[_T], default: _T2, *, apply_hook: bool = True
    ) -> Union[_T, _T2]: ...
    def get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        try:
            return self.extract(ref, apply_hook=apply_hook)
        except ContextPathError:
            if default is _MISSING:
                raise
            return default

    @overload
    async def async_get(
        self,
        ref: Ref[_T],
        *,
        apply_hook: bool = True,
    ) -> _T: ...
    @overload
    async def async_get(
        self,
        ref: Ref[_T],
        default: _T2,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]: ...
    async def async_get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        try:
            return await self.async_extract(ref, apply_hook=apply_hook)
        except ContextPathError:
            if default is _MISSING:
                raise
            return default

    def exists(self, ref: Ref[_T]) -> bool:
        if ref.key_path:
            raise ValueError(
                f"Ref.key_path must be empty for existence check (found {ref.key_path!r}). "
                "Use get() to check leaf value existence."
            )
        try:
            self._resolve(ref.parts)
            return True
        except ContextPathError:
            return False

    def keys(self, ref: Optional[Ref[_T]] = None) -> Iterable[str]:
        if ref is None:
            return self._root.keys()
        if ref.key_path:
            raise ValueError(
                f"Ref.key_path must be empty for listing keys (found {ref.key_path!r}). "
                "Context keys operation is structural and does not support leaf navigation."
            )
        element = self._resolve(ref.parts)
        if isinstance(element, ContextData):
            return element.keys()
        raise ContextPathError("Cannot list keys of a leaf value.")

    def to_context_data(self, ref: Optional[Ref[_T]] = None) -> ContextData:
        if ref is None:
            return self._root
        if ref.key_path:
            raise ValueError(
                f"Ref.key_path must be empty for converting to ContextData (found {ref.key_path!r}). "
                "ContextData conversion is structural and does not support leaf navigation."
            )
        val = self._resolve(ref.parts)
        if isinstance(val, ContextData):
            return val
        raise ContextPathError("Target is not a ContextData (container).")

    def to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        ref_tree = self._build_dict_ref_tree(ref)
        return self.extract(ref_tree, apply_hook=apply_hook)

    async def async_to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        ref_tree = self._build_dict_ref_tree(ref)
        return await self.async_extract(ref_tree, apply_hook=apply_hook)

    def _resolve(self, parts: Iterable[str]) -> Any:
        current: Any = self._root
        for p in parts:
            if not isinstance(current, ContextData):
                raise ContextPathError("Path blocked by leaf value.")
            try:
                current = current[p]
            except KeyError:
                raise ContextPathError(p) from None
        return current

    # --- Unified Modification Interface ---
    def mutate(
        self,
        *,
        updates: Optional[Mapping[Ref, Any]] = None,
        drops: Optional[Iterable[Ref]] = None,
        apply_hook: bool = True,
    ) -> "Context":
        """
        Apply a transaction-like set of modifications (updates and drops) atomically.

        Args:
            updates: A mapping of References to new values.
            drops: An iterable of References to remove.

        Returns:
            A new Context instance with the changes applied.
        """
        if not updates and not drops:
            return self

        updates = dict(updates) if updates else {}
        drops = set(drops) if drops else set()
        # Validate Ref.key_path is empty for all mutation operations
        invalid_updates = [r for r in updates if r.key_path]
        invalid_drops = [r for r in drops if r.key_path]
        if invalid_updates or invalid_drops:
            raise ValueError(
                f"Ref.key_path must be empty for mutation operations (found {invalid_updates!r} in updates "
                f"and {invalid_drops!r} in drops). Mutation on a specific key path is ambiguous; operate "
                "on the full path instead."
            )

        if apply_hook and self._hook:
            result = self._hook.on_mutate(ctx=self, updates=updates, drops=drops)
            updates, drops = result.updates, result.drops
        raw_updates = {r.parts: v for r, v in updates.items()}
        raw_drops = {r.parts for r in drops}

        # Delegate to the root ContextData
        new_root = self._root.mutate(raw_updates, raw_drops)
        # Edge Case: If the root itself resulted in MISSING (dropped), we reset to empty.
        if new_root is _MISSING:
            new_root = ContextData()
        return self._from_context_data(new_root, hook=self._hook)

    async def async_mutate(
        self,
        *,
        updates: Optional[Mapping[Ref, Any]] = None,
        drops: Optional[Iterable[Ref]] = None,
        apply_hook: bool = True,
    ) -> "Context":
        if not updates and not drops:
            return self

        updates = dict(updates) if updates else {}
        drops = set(drops) if drops else set()
        # Validate Ref.key_path is empty for all mutation operations
        invalid_updates = [r for r in updates if r.key_path]
        invalid_drops = [r for r in drops if r.key_path]
        if invalid_updates or invalid_drops:
            raise ValueError(
                f"Ref.key_path must be empty for mutation operations (found {invalid_updates!r} in updates "
                f"and {invalid_drops!r} in drops). Mutation on a specific key path is ambiguous; operate "
                "on the full path instead."
            )

        if apply_hook and self._hook:
            result = await self._hook.on_async_mutate(
                ctx=self, updates=updates, drops=drops
            )
            updates, drops = result.updates, result.drops
        raw_updates = {r.parts: v for r, v in updates.items()}
        raw_drops = {r.parts for r in drops}

        # Delegate to the root ContextData
        new_root = self._root.mutate(raw_updates, raw_drops)
        # Edge Case: If the root itself resulted in MISSING (dropped), we reset to empty.
        if new_root is _MISSING:
            new_root = ContextData()
        return self._from_context_data(new_root, hook=self._hook)

    # --- Convenience Interfaces ---
    def update(
        self, updates: Mapping[Ref, Any], *, apply_hook: bool = True
    ) -> "Context":
        """Batch update convenience interface."""
        return self.mutate(updates=updates, apply_hook=apply_hook)

    async def async_update(
        self, updates: Mapping[Ref, Any], *, apply_hook: bool = True
    ) -> "Context":
        return await self.async_mutate(updates=updates, apply_hook=apply_hook)

    def drop(self, refs: Iterable[Ref], *, apply_hook: bool = True) -> "Context":
        """Batch delete convenience interface."""
        return self.mutate(drops=refs, apply_hook=apply_hook)

    async def async_drop(
        self, refs: Iterable[Ref], *, apply_hook: bool = True
    ) -> "Context":
        return await self.async_mutate(drops=refs, apply_hook=apply_hook)

    def set(self, ref: Ref[_T], value: _T, *, apply_hook: bool = True) -> "Context":
        """Single set convenience interface."""
        return self.mutate(updates={ref: value}, apply_hook=apply_hook)

    async def async_set(
        self, ref: Ref[_T], value: _T, *, apply_hook: bool = True
    ) -> "Context":
        return await self.async_mutate(updates={ref: value}, apply_hook=apply_hook)

    def update_tree(
        self, ref_tree: Any, value_tree: Any, *, apply_hook: bool = True
    ) -> "Context":
        """
        Recursively update the context using a structure of references (ref_tree)
        and a matching structure of values (value_tree).

        Args:
            ref_tree: A nested structure (list, tuple, dict, MappingProxyType) where
                      leaves are references (Ref objects) that point to locations in the context.
            value_tree: A nested structure matching the shape of ref_tree, containing
                        the values to be updated at the corresponding references.

        Returns:
            A new Context instance with the updates applied.
        """
        updates = {
            ref: CTX_EVAL_ENGINE.get_element(value_tree, path)
            for path, ref in CTX_EVAL_ENGINE.iter_with_key_path(ref_tree)
        }
        return self.mutate(updates=updates, apply_hook=apply_hook)

    async def async_update_tree(
        self, ref_tree: Any, value_tree: Any, *, apply_hook: bool = True
    ) -> "Context":
        """
        Recursively update the context using a structure of references (ref_tree)
        and a matching structure of values (value_tree).

        Args:
            ref_tree: A nested structure (list, tuple, dict, MappingProxyType) where
                      leaves are references (Ref objects) that point to locations in the context.
            value_tree: A nested structure matching the shape of ref_tree, containing
                        the values to be updated at the corresponding references.

        Returns:
            A new Context instance with the updates applied.
        """
        updates = {
            ref: CTX_EVAL_ENGINE.get_element(value_tree, path)
            for path, ref in CTX_EVAL_ENGINE.iter_with_key_path(ref_tree)
        }
        return await self.async_mutate(updates=updates, apply_hook=apply_hook)

    def delete(self, ref: Ref[_T], *, apply_hook: bool = True) -> "Context":
        """Single delete convenience interface."""
        return self.mutate(drops=[ref], apply_hook=apply_hook)

    async def async_delete(self, ref: Ref[_T], *, apply_hook: bool = True) -> "Context":
        return await self.async_mutate(drops=[ref], apply_hook=apply_hook)

    def clear(self, ref: Ref[_T], *, apply_hook: bool = True) -> "Context":
        """
        Clear all contents under a reference but keep the path.
        Raises ContextPathError if the target is not a container (ContextData).
        """
        # 1. Validate target is a container
        val = self._resolve(ref.parts)
        if not isinstance(val, ContextData):
            raise ContextPathError(
                f"Cannot clear '{ref.path}': not a container (ContextData)."
            )

        # 2. Update with empty ContextData
        return self.mutate(updates={ref: ContextData()}, apply_hook=apply_hook)

    async def async_clear(self, ref: Ref[_T], *, apply_hook: bool = True) -> "Context":
        """
        Clear all contents under a reference but keep the path.
        Raises ContextPathError if the target is not a container (ContextData).
        """
        # 1. Validate target is a container
        val = self._resolve(ref.parts)
        if not isinstance(val, ContextData):
            raise ContextPathError(
                f"Cannot clear '{ref.path}': not a container (ContextData)."
            )

        # 2. Update with empty ContextData
        return await self.async_mutate(
            updates={ref: ContextData()}, apply_hook=apply_hook
        )


@dataclass(frozen=True, repr=False)
class ContextView(ContextElement):
    """
    Read-only view of a subtree within a Context.
    """

    _context: Context
    _parts: tuple[str, ...]

    def _adjust_ref(self, ref: Optional[Ref[_T]]) -> Ref[_T]:
        if ref is None:
            path = ".".join(self._parts)
            return Ref(path)
        new_parts = self._parts + ref.parts
        new_path = ".".join(new_parts)
        return Ref(new_path, key_path=ref.key_path, metadata=ref.metadata)

    def _adjust_ref_tree(self, ref_tree: Any) -> Any:
        def adjust(obj):
            if isinstance(obj, Ref):
                return self._adjust_ref(obj)
            return obj

        return CTX_EVAL_ENGINE.map(adjust, ref_tree)

    def extract(self, ref_tree: Any, *, apply_hook: bool = True, **kwargs) -> Any:
        return self._context.extract(
            self._adjust_ref_tree(ref_tree), apply_hook=apply_hook, **kwargs
        )

    async def async_extract(
        self, ref_tree: Any, *, apply_hook: bool = True, **kwargs
    ) -> Any:
        return await self._context.async_extract(
            self._adjust_ref_tree(ref_tree), apply_hook=apply_hook, **kwargs
        )

    def get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        return self._context.get(self._adjust_ref(ref), default, apply_hook=apply_hook)

    async def async_get(
        self,
        ref: Ref[_T],
        default: Union[_T2, _Missing] = _MISSING,
        *,
        apply_hook: bool = True,
    ) -> Union[_T, _T2]:
        return await self._context.async_get(
            self._adjust_ref(ref), default, apply_hook=apply_hook
        )

    def exists(self, ref: Ref[_T]) -> bool:
        return self._context.exists(self._adjust_ref(ref))

    def keys(self, ref: Optional[Ref[_T]] = None) -> Iterable[str]:
        return self._context.keys(self._adjust_ref(ref))

    def to_context_data(self, ref: Optional[Ref[_T]] = None) -> ContextData:
        return self._context.to_context_data(self._adjust_ref(ref))

    def to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        return self._context.to_dict(self._adjust_ref(ref), apply_hook=apply_hook)

    async def async_to_dict(
        self, ref: Optional[Ref[_T]] = None, *, apply_hook: bool = True
    ) -> dict[str, Any]:
        return await self._context.async_to_dict(
            self._adjust_ref(ref), apply_hook=apply_hook
        )


from .tree import CONTEXT_ENGINE, CTX_EVAL_ENGINE



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tree structure utilities for slyme (PyTree-like).

Designed to be lightweight, explicit, and instance-isolated.
"""

import types
from dataclasses import dataclass, field, replace
from collections.abc import Iterable, Callable, Iterator, Hashable, Mapping
from itertools import count
from typing import (
    Any,
    Union,
    Protocol,
    Optional,
    Literal,
)
from slyme.utils.registry import Registry, TypeRegistry

_EMPTY_MAPPING = types.MappingProxyType({})


@dataclass(frozen=True)
class PyTreeKey:
    """Base class for path elements."""

    def resolve(self, element: Any) -> Any:
        """
        Resolve the key against the given element to retrieve the child.
        Acts as the 'Getter' logic in KeyPath resolution.
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement resolve.")

    def codify(self, parent_expr: str) -> str:
        """
        Generate the Python code string to access this key from the parent expression.

        Args:
            parent_expr: The code expression of the parent container (e.g., 'tree', 'tree[0]').

        Returns:
            The new code expression (e.g., 'tree[0].value').
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement codify.")


@dataclass(frozen=True)
class SequenceKey(PyTreeKey):
    """Represents an index in a sequence (list, tuple)."""

    index: int

    def resolve(self, element: Any) -> Any:
        return element[self.index]

    def codify(self, parent_expr: str) -> str:
        return f"{parent_expr}[{self.index}]"


@dataclass(frozen=True)
class MappingKey(PyTreeKey):
    """Represents a key in a mapping (dict)."""

    key: Hashable

    def resolve(self, element: Any) -> Any:
        return element[self.key]

    def codify(self, parent_expr: str) -> str:
        return f"{parent_expr}[{repr(self.key)}]"


@dataclass(frozen=True)
class AttributeKey(PyTreeKey):
    """Represents an attribute name (object)."""

    name: str

    def resolve(self, element: Any) -> Any:
        return getattr(element, self.name)

    def codify(self, parent_expr: str) -> str:
        return f"{parent_expr}.{self.name}"


@dataclass(frozen=True)
class CallKey(PyTreeKey):
    """
    Represents a function call.

    If the function arguments are not hashable, this key falls back to identity semantics
    (object.__hash__ and self is other).
    """

    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Mapping[str, Any] = field(default_factory=lambda: _EMPTY_MAPPING)
    hash: int = field(init=False, repr=False)
    is_hashable: bool = field(init=False, repr=False)

    def __post_init__(self):
        # Convert kwargs to MappingProxyType if needed
        if not isinstance(self.kwargs, types.MappingProxyType):
            object.__setattr__(self, "kwargs", types.MappingProxyType(self.kwargs))

        try:
            # Sort kwargs for deterministic hashing
            # Assuming keys are strings as per standard kwargs
            kwargs_tuple = tuple(sorted(self.kwargs.items()))
            computed_hash = hash((self.args, kwargs_tuple))
            is_hashable = True
        except TypeError:
            computed_hash = object.__hash__(self)
            is_hashable = False

        object.__setattr__(self, "hash", computed_hash)
        object.__setattr__(self, "is_hashable", is_hashable)

    def __hash__(self) -> int:
        return self.hash

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CallKey):
            return False

        # If either is not hashable, fall back to identity
        if not self.is_hashable or not other.is_hashable:
            return self is other

        return self.args == other.args and self.kwargs == other.kwargs

    def resolve(self, element: Any) -> Any:
        return element(*self.args, **self.kwargs)

    def codify(self, parent_expr: str) -> str:
        args_str = [repr(arg) for arg in self.args]
        kwargs_str = [f"{k}={repr(v)}" for k, v in self.kwargs.items()]
        combined = ", ".join(args_str + kwargs_str)
        return f"{parent_expr}({combined})"


class KeyPathExpr:
    """
    A proxy object to build KeyPaths using natural syntax.

    Examples:
        P = KeyPathExpr()
        path = tuple(P.foo["bar"][0](1, a=2))
        # path is a tuple of (AttributeKey, MappingKey, MappingKey, CallKey)
    """

    __slots__ = ("_keys",)

    def __init__(self, keys: tuple[PyTreeKey, ...] = ()):
        object.__setattr__(self, "_keys", keys)

    def _raise_immutable(self, *args, **kwargs):
        raise TypeError(f"{type(self).__name__} is immutable.")

    __setattr__ = _raise_immutable
    __delattr__ = _raise_immutable

    def __getattribute__(self, name: str) -> Any:
        # Everything is treated as an AttributeKey.
        # Use object.__getattribute__ explicitly to get internal state when needed.
        current_keys = object.__getattribute__(self, "_keys")
        new_key = AttributeKey(name)
        return KeyPathExpr(current_keys + (new_key,))

    def __getitem__(self, key: Any) -> "KeyPathExpr":
        # Unified MappingKey for both sequence index and mapping key
        current_keys = object.__getattribute__(self, "_keys")
        new_key = MappingKey(key)
        return KeyPathExpr(current_keys + (new_key,))

    def __call__(self, *args: Any, **kwargs: Any) -> "KeyPathExpr":
        current_keys = object.__getattribute__(self, "_keys")
        new_key = CallKey(args=args, kwargs=kwargs)
        return KeyPathExpr(current_keys + (new_key,))

    def __iter__(self) -> Iterator[PyTreeKey]:
        return iter(object.__getattribute__(self, "_keys"))

    def __repr__(self) -> str:
        keys = object.__getattribute__(self, "_keys")
        return f"KeyPathExpr(keys={keys!r})"


# Global proxy object
P = KeyPathExpr()
# Type Alias for Path
KeyPath = tuple[PyTreeKey, ...]


@dataclass(frozen=True)
class PyTreeAux:
    """
    Auxiliary data required to reconstruct a container and track paths.

    Attributes:
        metadata: Custom data needed for unflattening (e.g., specific flags).
        children_keys: Optional tuple of keys corresponding to the children.
    """

    metadata: Mapping[str, Any] = field(default_factory=lambda: _EMPTY_MAPPING)
    children_keys: Optional[tuple[PyTreeKey, ...]] = None
    cls: Optional[type] = None

    def __post_init__(self):
        if not isinstance(self.metadata, types.MappingProxyType):
            object.__setattr__(self, "metadata", types.MappingProxyType(self.metadata))


class _FlattenFunc(Protocol):
    """
    Protocol for flattening a container.

    Returns:
        A tuple of (children_iterable, aux_data).
        - children_iterable: MUST yield pure child values.
        - aux_data: Contains metadata and optional keys for path tracking.
    """

    def __call__(self, element: Any, /) -> tuple[Iterable[Any], PyTreeAux]: ...


class _UnflattenFunc(Protocol):
    """
    Protocol for unflattening a container.
    """

    def __call__(self, children: Iterable[Any], tree_aux: PyTreeAux, /) -> Any: ...


@dataclass(frozen=True)
class _PyTreeHandler:
    flatten: _FlattenFunc
    unflatten: _UnflattenFunc


@dataclass(frozen=True)
class TraverseAux:
    """
    Auxiliary data during traversal.
    """

    parent: Any
    key_path: KeyPath


class _IsLeafFunc(Protocol):
    """
    Protocol for functions that determine if an element is a leaf.
    """

    def __call__(self, element: Any, traverse_aux: TraverseAux, /) -> bool: ...


class _ResolverFunc(Protocol):
    """
    Protocol for dynamic handler resolution.
    Accepts an auxiliary object.
    """

    def __call__(
        self, element: Any, traverse_aux: TraverseAux, /
    ) -> Union[_PyTreeHandler, None]: ...


class _LeafSinkFunc(Protocol):
    """Internal protocol for collecting leaves."""

    def __call__(self, leaf: Any, traverse_aux: TraverseAux, /) -> None: ...


@dataclass(frozen=True)
class PyTreeDef:
    """Base class for tree definitions."""

    def unflatten(self, leaves: Iterable[Any]) -> Any:
        """
        Public API: Reconstruct the object from this structure and leaves.
        """
        leaves_iter = iter(leaves)
        element = self._build(leaves_iter)

        try:
            next(leaves_iter)
            raise ValueError("Too many leaves provided for this tree structure.")
        except StopIteration:
            pass

        return element

    def _build(self, leaves_iter: Iterator[Any]) -> Any:
        """Internal recursive driver."""
        raise NotImplementedError


@dataclass(frozen=True)
class LeafDef(PyTreeDef):
    def _build(self, leaves_iter: Iterator[Any]) -> Any:
        try:
            return next(leaves_iter)
        except StopIteration:
            raise ValueError("Too few leaves provided for this tree structure.")


@dataclass(frozen=True)
class ContainerDef(PyTreeDef):
    cls: type
    tree_aux: PyTreeAux
    children_defs: tuple[PyTreeDef, ...]
    unflatten_func: _UnflattenFunc = field(compare=False, repr=False)

    def _build(self, leaves_iter: Iterator[Any]) -> Any:
        children = tuple(child._build(leaves_iter) for child in self.children_defs)
        return self.unflatten_func(children, self.tree_aux)


class PyTreeEngine:
    """
    A PyTree Engine that defines PyTree Operations.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        strict_registration: bool = True,
        allow_inheritance: bool = True,
        register_defaults: bool = True,
    ) -> None:
        self.name = repr(self) if name is None else name
        self.allow_inheritance = allow_inheritance

        # 1. Type-based Registry (O(1) lookup, Middle Priority)
        self._registry: TypeRegistry[Any, _PyTreeHandler] = TypeRegistry(
            f"PyTreeTypeRegistry<{self.name}>", strict=strict_registration
        )

        # 2. Dynamic Resolvers
        # Pre-resolvers: Checked BEFORE Type Registry (High Priority)
        self._pre_resolvers: list[_ResolverFunc] = []
        # Post-resolvers: Checked AFTER Type Registry (Low Priority)
        self._post_resolvers: list[_ResolverFunc] = []

        if register_defaults:
            self._register_defaults()

    def register(
        self,
        cls: type,
        flatten_func: _FlattenFunc,
        unflatten_func: _UnflattenFunc,
        strict: bool = True,
    ) -> None:
        """
        Register a custom type handler into the core TypeRegistry.
        This follows the explicit-is-better-than-implicit philosophy.
        """
        handler = _PyTreeHandler(flatten=flatten_func, unflatten=unflatten_func)
        self._registry.register(handler, key=cls, strict=strict)

    def register_resolver(
        self, resolver: _ResolverFunc, *, priority: Literal["pre", "post"] = "post"
    ) -> None:
        """
        Register a dynamic resolver function.

        Args:
            resolver: A function taking an element and returning a Handler or None.
            priority:
                - 'pre': Checked BEFORE the core TypeRegistry. Used to override
                  default behaviors or intercept specific instances.
                - 'post': Checked AFTER the core TypeRegistry. Used for generic
                  fallbacks (e.g., Dataclasses, Protocol checks).
        """
        if priority == "pre":
            self._pre_resolvers.append(resolver)
        elif priority == "post":
            self._post_resolvers.append(resolver)
        else:
            raise ValueError(f"Invalid priority: {priority}. Must be 'pre' or 'post'.")

    def _register_defaults(self) -> None:
        """Register standard python containers."""
        # Tuple
        self.register(
            tuple,
            lambda x: (iter(x), PyTreeAux()),
            lambda children, _: tuple(children),
        )
        # List
        self.register(
            list,
            lambda x: (iter(x), PyTreeAux()),
            lambda children, _: list(children),
        )
        # Dict
        self.register(dict, flatten_dict, unflatten_dict)

    def _lookup_handler(
        self, element: Any, traverse_aux: TraverseAux
    ) -> Union[_PyTreeHandler, None]:
        """
        Resolve handler via:
        1. Pre-resolvers (High Priority)
        2. Type Registry (Core Priority)
        3. Post-resolvers (Fallback Priority)
        """
        # 1. Try Pre-resolvers
        for resolver in self._pre_resolvers:
            handler = resolver(element, traverse_aux)
            if handler is not None:
                return handler

        # 2. Try Type Registry
        cls = type(element)
        handler = None
        if self.allow_inheritance:
            handler = self._registry.lookup(cls, default=None)
        else:
            handler = self._registry.get(cls, None)

        if handler is not None:
            return handler

        # 3. Try Post-resolvers
        for resolver in self._post_resolvers:
            handler = resolver(element, traverse_aux)
            if handler is not None:
                return handler

        return None

    def flatten(
        self,
        tree: Any,
        *,
        is_leaf: Optional[_IsLeafFunc] = None,
    ) -> tuple[list[Any], "PyTreeDef"]:
        """
        Flatten a tree into a list of leaves and a structure definition.
        """
        leaves: list[Any] = []

        def _sink(leaf: Any, traverse_aux: TraverseAux) -> None:
            leaves.append(leaf)

        initial_traverse_aux = TraverseAux(parent=None, key_path=())
        treedef = self._traverse(tree, initial_traverse_aux, _sink, is_leaf)
        return leaves, treedef

    def flatten_with_key_path(
        self,
        tree: Any,
        *,
        is_leaf: Optional[_IsLeafFunc] = None,
    ) -> tuple[list[tuple[KeyPath, Any]], "PyTreeDef"]:
        """
        Flatten a tree into a list of (key_path, leaf) tuples and a structure definition.
        """
        leaves_with_path: list[tuple[KeyPath, Any]] = []

        def _sink(leaf: Any, traverse_aux: TraverseAux) -> None:
            leaves_with_path.append((traverse_aux.key_path, leaf))

        initial_traverse_aux = TraverseAux(parent=None, key_path=())
        treedef = self._traverse(tree, initial_traverse_aux, _sink, is_leaf)
        return leaves_with_path, treedef

    def iter(
        self,
        tree: Any,
        *,
        is_leaf: Optional[_IsLeafFunc] = None,
    ) -> Iterator[Any]:
        """
        Iterate over leaves of a tree without creating a PyTreeDef.
        """
        initial_traverse_aux = TraverseAux(parent=None, key_path=())
        yield from self._traverse_iter(
            tree, initial_traverse_aux, is_leaf, with_key_path=False
        )

    def iter_with_key_path(
        self,
        tree: Any,
        *,
        is_leaf: Optional[_IsLeafFunc] = None,
    ) -> Iterator[tuple[KeyPath, Any]]:
        """
        Iterate over (key_path, leaf) tuples of a tree without creating a PyTreeDef.
        """
        initial_traverse_aux = TraverseAux(parent=None, key_path=())
        yield from self._traverse_iter(
            tree, initial_traverse_aux, is_leaf, with_key_path=True
        )

    def _prepare_element(
        self,
        element: Any,
        traverse_aux: TraverseAux,
        is_leaf: Optional[_IsLeafFunc],
    ) -> tuple[bool, Optional[_PyTreeHandler], Iterable[Any], Iterator[Any], PyTreeAux]:
        """
        Helper to check if an element should be flattened and prepare iterators.
        Returns: (should_flatten, handler, children_iter, keys_iter, tree_aux)
        """
        # should_flatten check
        should_flatten = is_leaf is None or not is_leaf(element, traverse_aux)
        handler = None
        if should_flatten:
            handler = self._lookup_handler(element, traverse_aux)
            should_flatten = handler is not None

        if not should_flatten:
            # Return defaults for non-flattenable
            return False, None, [], iter([]), PyTreeAux()

        children_iter, tree_aux = handler.flatten(element)

        # Auto fill tree_aux info
        # TODO: Maybe refactor this into a function when the
        # auto fill logic grows.
        if tree_aux.cls is None:
            tree_aux = replace(tree_aux, cls=type(element))

        # Resolve Keys for Path Tracking.
        if tree_aux.children_keys is not None:
            keys_iter = iter(tree_aux.children_keys)
        else:
            # Default fallback: Generate SequenceKey for indices.
            keys_iter = (SequenceKey(i) for i in count())

        return True, handler, children_iter, keys_iter, tree_aux

    def _traverse(
        self,
        element: Any,
        traverse_aux: TraverseAux,
        leaf_sink: _LeafSinkFunc,
        is_leaf: Optional[_IsLeafFunc],
    ) -> PyTreeDef:
        """Recursive core for traversal."""
        should_flatten, handler, children_iter, keys_iter, tree_aux = (
            self._prepare_element(element, traverse_aux, is_leaf)
        )

        if should_flatten:
            # Recursively map children.
            child_defs = []
            for child in children_iter:
                try:
                    key = next(keys_iter)
                except StopIteration:
                    # Should not happen if aux.key_path matches children length.
                    raise ValueError(
                        f"Not enough keys provided in TreeAux for container {type(element)}"
                    )

                child_traverse_aux = TraverseAux(
                    parent=element, key_path=traverse_aux.key_path + (key,)
                )
                child_def = self._traverse(
                    child,
                    child_traverse_aux,
                    leaf_sink,
                    is_leaf,
                )
                child_defs.append(child_def)

            return ContainerDef(
                type(element), tree_aux, tuple(child_defs), handler.unflatten
            )
        else:
            # Leaf.
            leaf_sink(element, traverse_aux)
            return LeafDef()

    def _traverse_iter(
        self,
        element: Any,
        traverse_aux: TraverseAux,
        is_leaf: Optional[_IsLeafFunc],
        with_key_path: bool,
    ) -> Iterator[Any]:
        """Recursive core for iterator traversal."""
        should_flatten, _, children_iter, keys_iter, _ = self._prepare_element(
            element, traverse_aux, is_leaf
        )

        if should_flatten:
            for child in children_iter:
                try:
                    key = next(keys_iter)
                except StopIteration:
                    # Should not happen if aux.key_path matches children length.
                    raise ValueError(
                        f"Not enough keys provided in TreeAux for container {type(element)}"
                    )

                child_traverse_aux = TraverseAux(
                    parent=element, key_path=traverse_aux.key_path + (key,)
                )
                yield from self._traverse_iter(
                    child,
                    child_traverse_aux,
                    is_leaf,
                    with_key_path,
                )
        else:
            # Leaf.
            if with_key_path:
                yield (traverse_aux.key_path, element)
            else:
                yield element

    @staticmethod
    def unflatten(treedef: "PyTreeDef", leaves: Iterable[Any]) -> Any:
        """
        Reconstruct the tree from a structure definition and a list of leaves.
        """
        return treedef.unflatten(leaves)

    def map(
        self,
        func: Callable[..., Any],
        tree: Any,
        *,
        is_leaf: Optional[_IsLeafFunc] = None,
    ) -> Any:
        """Apply func to every leaf in the tree."""
        leaves, treedef = self.flatten(tree, is_leaf=is_leaf)
        new_leaves = [func(leaf) for leaf in leaves]
        return self.unflatten(treedef, new_leaves)

    @staticmethod
    def get_element(tree: Any, key_path: KeyPath) -> Any:
        """
        Retrieve an element from the tree using a specific key_path (Runtime KeyPath Resolution).
        """
        current = tree
        for i, key in enumerate(key_path):
            if not isinstance(key, PyTreeKey):
                raise TypeError(
                    f"Invalid key_path key at index {i}: expected PyTreeKey, got {type(key)}."
                )
            current = key.resolve(current)
        return current

    @staticmethod
    def codify_key_path(key_path: KeyPath, root_name: str = "$") -> str:
        """
        Generate the Python code string corresponding to the key_path.
        """
        expr = root_name
        for key in key_path:
            expr = key.codify(expr)
        return expr


# Global registry to manage Tree instances.
PYTREE_ENGINE_REGISTRY: Registry[PyTreeEngine] = Registry("pytree_engine")

from .common import flatten_dict, unflatten_dict



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core node module, consolidating base definitions and functional APIs.
"""

import types
import inspect
from abc import ABC, abstractmethod
from enum import Enum
from functools import partial, update_wrapper
from typing import (
    TypeVar,
    Callable,
    Any,
    Union,
    overload,
    Mapping,
    Optional,
    Generic,
    Iterable,
    Sequence,
    Awaitable,
)
from typing_extensions import ParamSpec, Concatenate, Self
from slyme.utils.exception import enrich_exception
from slyme.context import Context
from .exception import (
    NodeTerminate,
    NodeExceptionRecord,
    NodeException,
    ExpressionExceptionRecord,
    WrapperExceptionRecord,
)
from .signature import (
    Spec,
    process_kwargs,
    analyze_signature,
    resolve_arguments,
)

__all__ = [
    "Config",
    "node",
    "expression",
    "wrapper",
    "NodeElement",
    "Node",
    "NodeDef",
    "NodeExec",
    "Expression",
    "ExpressionDef",
    "ExpressionExec",
    "Wrapper",
    "WrapperDef",
    "WrapperExec",
    "AsyncNode",
    "AsyncNodeDef",
    "AsyncNodeExec",
    "AsyncExpression",
    "AsyncExpressionDef",
    "AsyncExpressionExec",
    "AsyncWrapper",
    "AsyncWrapperDef",
    "AsyncWrapperExec",
    "NODE_ENGINE",
]

_P = ParamSpec("_P")
_R = TypeVar("_R")

# Update: NodeFunc now must return Context to align with the chain protocol.
NodeFunc = Callable[Concatenate[Context, _P], Context]
ExpressionFunc = Callable[Concatenate[Context, _P], _R]
# Wrapper definition: (ctx, wrapped, call_next, ...args) -> Context
# call_next definition: (Context) -> Context
WrapperFunc = Callable[
    Concatenate[Context, "Node", Callable[[Context], Context], _P], Context
]
AsyncNodeFunc = Callable[Concatenate[Context, _P], Awaitable[Context]]
AsyncExpressionFunc = Callable[Concatenate[Context, _P], Awaitable[_R]]
AsyncWrapperFunc = Callable[
    Concatenate[Context, "AsyncNode", Callable[[Context], Awaitable[Context]], _P],
    Awaitable[Context],
]
_Missing = Enum("_Missing", ["MARK"])
_MISSING = _Missing.MARK


class Config:
    check_return_type: bool = True


def _ensure_context_return(func: Callable[_P, Any]) -> Callable[_P, Context]:
    """
    Wrap a callable to ensure it returns a Context object.
    """

    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> Context:
        result = func(*args, **kwargs)
        if not isinstance(result, Context):
            raise TypeError(
                f"Node execution return type mismatch. "
                f"Expected 'Context', but got '{type(result).__name__}'. "
                f"Function: {func}"
            )
        return result

    return wrapper


def _ensure_async_context_return(
    func: Callable[_P, Awaitable[Any]],
) -> Callable[_P, Awaitable[Context]]:
    """
    Wrap an async callable to ensure it returns a Context object.
    """

    async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> Context:
        result = await func(*args, **kwargs)
        if not isinstance(result, Context):
            raise TypeError(
                f"Node execution return type mismatch. "
                f"Expected 'Context', but got '{type(result).__name__}'. "
                f"Function: {func}"
            )
        return result

    return wrapper


def _prepare_eval(
    specs: Mapping[str, Spec], kwargs: Mapping[str, Any]
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Split kwargs into static values and dynamic evaluators based on specs.
    """
    raw_kwargs = {}
    eval_kwargs = {}
    for key, value in kwargs.items():
        if specs[key].should_eval(value):
            eval_kwargs[key] = value
        else:
            raw_kwargs[key] = value
    return raw_kwargs, eval_kwargs


class _DefMixin:
    """
    Mixin for mutable definition classes.
    """

    _specs: Mapping[str, Spec]
    _kwargs: dict[str, Any]

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise TypeError(
            f"{type(self).__name__} is a definition and not callable. "
            f"Please call `.prepare()` to obtain an Exec instance first."
        )

    def __setitem__(self, key: str, value: Any) -> None:
        if key not in self._specs:
            raise KeyError(
                f"Invalid key '{key}'. Parameters must be defined in the specs."
            )
        spec_obj = self._specs[key]
        with enrich_exception(f"for parameter '{key}'"):
            value = spec_obj._build(value)
        self._kwargs[key] = value

    def update(self, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            self[k] = v

    def __delattr__(self, name: str) -> None:
        raise AttributeError(
            f"Cannot delete attribute '{name}' on {type(self).__name__}"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError(
            f"Cannot set attribute '{name}' on {type(self).__name__}. "
            "Internal structure is protected."
        )


class _ExecMixin:
    """
    Mixin for immutable execution classes.
    """

    def __setitem__(self, key: str, value: Any) -> None:
        raise TypeError(
            f"{type(self).__name__} is immutable and does not support item assignment."
        )

    def __delattr__(self, name: str) -> None:
        raise TypeError(f"{type(self).__name__} is immutable.")

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(f"{type(self).__name__} is immutable.")


# Node Family
class NodeElement(ABC):
    """
    Base class for all node-related entities.
    Abstract definition that separates API from implementation.
    """

    _func: Callable
    _specs: Mapping[str, Spec]
    _kwargs: Mapping[str, Any]

    @abstractmethod
    def prepare(self) -> "NodeElement":
        """
        Prepare the node for execution.
        Converts the Definition structure into an Execution structure.
        """
        pass

    def __getitem__(self, key: str) -> Any:
        return self._kwargs[key]

    def __repr__(self) -> str:
        return get_render_string(self)

    def extra_repr(self) -> str:
        return ""

    def type_repr(self) -> str:
        return f"{self._func.__name__}<{self.__class__.__name__}>"


class BaseNode(NodeElement):
    pass


class Node(BaseNode):
    """
    Abstract base class for NodeDef and NodeExec.
    """

    _func: NodeFunc
    wrappers: Sequence["Wrapper"]  # Changed: Union[...] -> Sequence

    @abstractmethod
    def __call__(self, ctx: Context) -> Context:
        pass


class NodeDef(_DefMixin, Node):
    """
    Mutable definition of a Node. Allows modification during build time.
    """

    _kwargs: dict[str, Any]
    wrappers: list["Wrapper"]

    def __init__(
        self,
        /,
        *,
        func: NodeFunc,
        specs: Mapping[str, Spec],
        wrappers: Optional[Iterable["Wrapper"]] = None,
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "wrappers", list(wrappers) if wrappers else [])
        object.__setattr__(self, "_kwargs", kwargs)

    def add_wrappers(self, *wrappers: "Wrapper") -> Self:
        self.wrappers.extend(wrappers)
        return self

    def prepare(self) -> "NodeExec":
        # Use the specialized NODE_PREPARE_PYTREE_ENGINE to perform a deep transform
        # of the structure (List -> Tuple, Dict -> MappingProxy, Def -> Exec).
        # We map strict identity because the transformation happens in the 'unflatten' phase
        # of the registered types in NODE_PREPARE_PYTREE_ENGINE.
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "wrappers":
            object.__setattr__(self, name, value)
        else:
            super().__setattr__(name, value)


class NodeExec(_ExecMixin, Node):
    """
    Immutable execution version of a Node.
    """

    # composed_func signature: (Context) -> Context
    _prepared_func: Callable[[Context], Context]
    wrappers: tuple["Wrapper", ...]

    def __init__(
        self,
        /,
        *,
        func: NodeFunc,
        specs: Mapping[str, Spec],
        wrappers: Iterable["Wrapper"],
        kwargs: Mapping[str, Any],
    ):
        wrappers = tuple(wrappers)
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "wrappers", wrappers)
        object.__setattr__(self, "_kwargs", kwargs)
        # --- Composition Logic (Onion Model) ---
        # 1. Inner Core: Bind kwargs to the user function.
        # Signature: (Context) -> Context
        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            chain: Callable[[Context], Context] = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            def chain(ctx: Context) -> Context:
                return func(ctx, **raw_kwargs, **execute_eval_plan(ctx, eval_plan))

        if Config.check_return_type:
            chain = _ensure_context_return(chain)

        # 2. Build the middleware chain.
        # Wrappers are applied from inside out (reversed order of list).
        for wrapper in reversed(wrappers):
            # wrapper is WrapperExec which is callable: (ctx, wrapped, call_next) -> Context
            # We partially apply `wrapped` (self) and `call_next` (current chain head)
            # to create the new chain head: (Context) -> Context
            chain = partial(wrapper, wrapped=self, call_next=chain)
        object.__setattr__(self, "_prepared_func", chain)

    def prepare(self) -> Self:
        return self

    def __call__(self, ctx: Context) -> Context:
        try:
            # Execute the pre-composed chain
            return self._prepared_func(ctx)
        # Node Interrupts
        except (NodeTerminate, ExpressionExceptionRecord) as e:
            if e.source_node is None:
                e.source_node = self
            raise
        # Direct Node Exceptions
        except NodeException:
            raise
        # General Exceptions
        except Exception as e:
            raise NodeExceptionRecord(exception_node=self, exception=e) from e


class AsyncNode(BaseNode):
    """
    Abstract base class for AsyncNodeDef and AsyncNodeExec.
    """

    _func: AsyncNodeFunc
    wrappers: Sequence["AsyncWrapper"]

    @abstractmethod
    async def __call__(self, ctx: Context) -> Context:
        pass


class AsyncNodeDef(_DefMixin, AsyncNode):
    """
    Mutable definition of an AsyncNode.
    """

    _kwargs: dict[str, Any]
    wrappers: list["AsyncWrapper"]

    def __init__(
        self,
        /,
        *,
        func: AsyncNodeFunc,
        specs: Mapping[str, Spec],
        wrappers: Optional[Iterable["AsyncWrapper"]] = None,
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "wrappers", list(wrappers) if wrappers else [])
        object.__setattr__(self, "_kwargs", kwargs)

    def add_wrappers(self, *wrappers: "AsyncWrapper") -> Self:
        self.wrappers.extend(wrappers)
        return self

    def prepare(self) -> "AsyncNodeExec":
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "wrappers":
            object.__setattr__(self, name, value)
        else:
            super().__setattr__(name, value)


class AsyncNodeExec(_ExecMixin, AsyncNode):
    """
    Immutable execution version of an AsyncNode.
    """

    _prepared_func: Callable[[Context], Awaitable[Context]]
    wrappers: tuple["AsyncWrapper", ...]

    def __init__(
        self,
        /,
        *,
        func: AsyncNodeFunc,
        specs: Mapping[str, Spec],
        wrappers: Iterable["AsyncWrapper"],
        kwargs: Mapping[str, Any],
    ):
        wrappers = tuple(wrappers)
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "wrappers", wrappers)
        object.__setattr__(self, "_kwargs", kwargs)

        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            chain = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            async def chain(ctx: Context) -> Context:
                return await func(
                    ctx, **raw_kwargs, **await async_execute_eval_plan(ctx, eval_plan)
                )

        if Config.check_return_type:
            chain = _ensure_async_context_return(chain)

        for wrapper in reversed(wrappers):
            chain = partial(wrapper, wrapped=self, call_next=chain)
        object.__setattr__(self, "_prepared_func", chain)

    def prepare(self) -> Self:
        return self

    async def __call__(self, ctx: Context) -> Context:
        try:
            return await self._prepared_func(ctx)
        except (NodeTerminate, ExpressionExceptionRecord) as e:
            if e.source_node is None:
                e.source_node = self
            raise
        except NodeException:
            raise
        except Exception as e:
            raise NodeExceptionRecord(exception_node=self, exception=e) from e


class BaseExpression(NodeElement):
    pass


class Expression(BaseExpression, Generic[_R]):
    """
    Abstract base class for ExpressionDef and ExpressionExec.
    """

    _func: ExpressionFunc

    @abstractmethod
    def __call__(self, ctx: Context) -> _R:
        pass


class ExpressionDef(_DefMixin, Expression[_R]):
    """
    Mutable definition of an Expression.
    """

    _kwargs: dict[str, Any]  # Override: Mapping -> dict

    def __init__(
        self,
        /,
        *,
        func: ExpressionFunc,
        specs: Mapping[str, Spec],
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

    def prepare(self) -> "ExpressionExec[_R]":
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)


class ExpressionExec(_ExecMixin, Expression[_R]):
    """
    Immutable execution version of an Expression.
    """

    # composed_func signature: (Context) -> _R
    _prepared_func: Callable[[Context], _R]

    def __init__(
        self,
        /,
        *,
        func: ExpressionFunc,
        specs: Mapping[str, Spec],
        kwargs: Mapping[str, Any],
    ):
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)
        # Optimization: Pre-bind kwargs using partial
        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            prepared_func = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            def prepared_func(ctx: Context) -> _R:
                return func(ctx, **raw_kwargs, **execute_eval_plan(ctx, eval_plan))

        object.__setattr__(self, "_prepared_func", prepared_func)

    def prepare(self) -> Self:
        return self

    def __call__(self, ctx: Context) -> _R:
        try:
            return self._prepared_func(ctx)
        except NodeException:
            raise
        except Exception as e:
            raise ExpressionExceptionRecord(exception_node=self, exception=e) from e


class AsyncExpression(BaseExpression, Generic[_R]):
    """
    Abstract base class for AsyncExpressionDef and AsyncExpressionExec.
    """

    _func: AsyncExpressionFunc

    @abstractmethod
    async def __call__(self, ctx: Context) -> _R:
        pass


class AsyncExpressionDef(_DefMixin, AsyncExpression[_R]):
    """
    Mutable definition of an AsyncExpression.
    """

    _kwargs: dict[str, Any]

    def __init__(
        self,
        /,
        *,
        func: AsyncExpressionFunc,
        specs: Mapping[str, Spec],
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

    def prepare(self) -> "AsyncExpressionExec[_R]":
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)


class AsyncExpressionExec(_ExecMixin, AsyncExpression[_R]):
    """
    Immutable execution version of an AsyncExpression.
    """

    _prepared_func: Callable[[Context], Awaitable[_R]]

    def __init__(
        self,
        /,
        *,
        func: AsyncExpressionFunc,
        specs: Mapping[str, Spec],
        kwargs: Mapping[str, Any],
    ):
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            prepared_func = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            async def prepared_func(ctx: Context) -> _R:
                return await func(
                    ctx, **raw_kwargs, **await async_execute_eval_plan(ctx, eval_plan)
                )

        object.__setattr__(self, "_prepared_func", prepared_func)

    def prepare(self) -> Self:
        return self

    async def __call__(self, ctx: Context) -> _R:
        try:
            return await self._prepared_func(ctx)
        except NodeException:
            raise
        except Exception as e:
            raise ExpressionExceptionRecord(exception_node=self, exception=e) from e


class BaseWrapper(NodeElement):
    pass


class Wrapper(BaseWrapper):
    """
    Abstract base class for WrapperDef and WrapperExec.
    """

    _func: WrapperFunc

    @abstractmethod
    def __call__(
        self,
        ctx: Context,
        wrapped: Node,
        call_next: Callable[[Context], Context],
    ) -> Context:
        pass


class WrapperDef(_DefMixin, Wrapper):
    """
    Mutable definition of a Wrapper.
    """

    _kwargs: dict[str, Any]  # Override: Mapping -> dict

    def __init__(
        self,
        /,
        *,
        func: WrapperFunc,
        specs: Mapping[str, Spec],
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

    def prepare(self) -> "WrapperExec":
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)


class WrapperExec(_ExecMixin, Wrapper):
    """
    Immutable execution version of a Wrapper.
    """

    # composed_func signature: (ctx, wrapped, call_next) -> Context
    _prepared_func: Callable[
        [Context, Node, Callable[[Context], Context]],
        Context,
    ]

    def __init__(
        self,
        /,
        *,
        func: WrapperFunc,
        specs: Mapping[str, Spec],
        kwargs: Mapping[str, Any],
    ):
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)
        # Optimization: Pre-bind kwargs using partial
        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            prepared_func = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            def prepared_func(
                ctx: Context,
                wrapped: Node,
                call_next: Callable[[Context], Context],
            ) -> Context:
                return func(
                    ctx,
                    wrapped,
                    call_next,
                    **raw_kwargs,
                    **execute_eval_plan(ctx, eval_plan),
                )

        if Config.check_return_type:
            prepared_func = _ensure_context_return(prepared_func)
        object.__setattr__(self, "_prepared_func", prepared_func)

    def prepare(self) -> Self:
        return self

    def __call__(
        self,
        ctx: Context,
        wrapped: Node,
        call_next: Callable[[Context], Context],
    ) -> Context:
        try:
            return self._prepared_func(ctx, wrapped, call_next)
        except NodeException:
            raise
        except Exception as e:
            raise WrapperExceptionRecord(
                exception_node=self, wrapped_node=wrapped, exception=e
            ) from e


class AsyncWrapper(BaseWrapper):
    """
    Abstract base class for AsyncWrapperDef and AsyncWrapperExec.
    """

    _func: AsyncWrapperFunc

    @abstractmethod
    async def __call__(
        self,
        ctx: Context,
        wrapped: AsyncNode,
        call_next: Callable[[Context], Awaitable[Context]],
    ) -> Context:
        pass


class AsyncWrapperDef(_DefMixin, AsyncWrapper):
    """
    Mutable definition of an AsyncWrapper.
    """

    _kwargs: dict[str, Any]

    def __init__(
        self,
        /,
        *,
        func: AsyncWrapperFunc,
        specs: Mapping[str, Spec],
        kwargs: dict[str, Any],
    ):
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

    def prepare(self) -> "AsyncWrapperExec":
        return NODE_PREPARE_ENGINE.map(lambda x: x, self)


class AsyncWrapperExec(_ExecMixin, AsyncWrapper):
    """
    Immutable execution version of an AsyncWrapper.
    """

    _prepared_func: Callable[
        [Context, Node, Callable[[Context], Awaitable[Context]]],
        Awaitable[Context],
    ]

    def __init__(
        self,
        /,
        *,
        func: AsyncWrapperFunc,
        specs: Mapping[str, Spec],
        kwargs: Mapping[str, Any],
    ):
        if not isinstance(kwargs, types.MappingProxyType):
            kwargs = types.MappingProxyType(kwargs)
        object.__setattr__(self, "_func", func)
        object.__setattr__(self, "_specs", specs)
        object.__setattr__(self, "_kwargs", kwargs)

        raw_kwargs, eval_kwargs = _prepare_eval(specs, kwargs)

        if not eval_kwargs:
            prepared_func = partial(func, **raw_kwargs)
        else:
            eval_plan = prepare_eval_plan(eval_kwargs)

            async def prepared_func(
                ctx: Context,
                wrapped: AsyncNode,
                call_next: Callable[[Context], Awaitable[Context]],
            ) -> Context:
                return await func(
                    ctx,
                    wrapped,
                    call_next,
                    **raw_kwargs,
                    **await async_execute_eval_plan(ctx, eval_plan),
                )

        if Config.check_return_type:
            prepared_func = _ensure_async_context_return(prepared_func)
        object.__setattr__(self, "_prepared_func", prepared_func)

    def prepare(self) -> Self:
        return self

    async def __call__(
        self,
        ctx: Context,
        wrapped: AsyncNode,
        call_next: Callable[[Context], Awaitable[Context]],
    ) -> Context:
        try:
            return await self._prepared_func(ctx, wrapped, call_next)
        except NodeException:
            raise
        except Exception as e:
            raise WrapperExceptionRecord(
                exception_node=self, wrapped_node=wrapped, exception=e
            ) from e


# Functional Factory & Decorators
class BaseFactory:
    _func: Callable

    @property
    def func(self) -> Callable:
        return self._func

    def __repr__(self) -> str:
        return f"<{type(self).__name__} of {self._func.__name__}>"


class NodeFactory(BaseFactory, Generic[_P]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> NodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> NodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> NodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> NodeDef: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> NodeDef: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> NodeDef:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        # Note: wrappers are intentionally omitted to avoid parameter conflict.
        # Users should use .add_wrappers() explicitly.
        return NodeDef(func=self._func, specs=self._specs, kwargs=final_kwargs)


class ExpressionFactory(BaseFactory, Generic[_P, _R]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> ExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> ExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> ExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> ExpressionDef[_R]: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> ExpressionDef[_R]: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> ExpressionDef[_R]:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        return ExpressionDef(func=self._func, specs=self._specs, kwargs=final_kwargs)


class WrapperFactory(BaseFactory, Generic[_P]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> WrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> WrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> WrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> WrapperDef: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> WrapperDef: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> WrapperDef:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        return WrapperDef(func=self._func, specs=self._specs, kwargs=final_kwargs)


class AsyncNodeFactory(BaseFactory, Generic[_P]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncNodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncNodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncNodeDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncNodeDef: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncNodeDef: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncNodeDef:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        # Note: wrappers are intentionally omitted to avoid parameter conflict.
        # Users should use .add_wrappers() explicitly.
        return AsyncNodeDef(func=self._func, specs=self._specs, kwargs=final_kwargs)


class AsyncExpressionFactory(BaseFactory, Generic[_P, _R]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncExpressionDef[_R]: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncExpressionDef[_R]: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncExpressionDef[_R]: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncExpressionDef[_R]:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        return AsyncExpressionDef(
            func=self._func, specs=self._specs, kwargs=final_kwargs
        )


class AsyncWrapperFactory(BaseFactory, Generic[_P]):
    def __init__(
        self,
        func: Callable,
        specs: Mapping[str, Spec],
        signature: inspect.Signature,
    ):
        update_wrapper(self, func)
        self._func = func
        self._specs = specs
        self.__signature__ = signature

    @overload
    def __call__(
        self,
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncWrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncWrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncWrapperDef: ...
    @overload
    def __call__(
        self,
        scope1: Mapping[str, Any],
        scope2: Mapping[str, Any],
        scope3: Mapping[str, Any],
        /,
        *_: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncWrapperDef: ...
    @overload
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncWrapperDef: ...
    def __call__(self, *scopes: Any, **kwargs: Any) -> AsyncWrapperDef:
        """
        Create the node instance by resolving parameters from scopes and overrides.
        """
        resolved_kwargs = resolve_arguments(self._specs, scopes, kwargs)
        # Process kwargs
        with enrich_exception(f"for '{self._func.__name__}'"):
            final_kwargs = process_kwargs(self._specs, resolved_kwargs)
        return AsyncWrapperDef(func=self._func, specs=self._specs, kwargs=final_kwargs)


def _node(func: NodeFunc[_P], /, *, resolve_type_hints: bool) -> NodeFactory[_P]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    if len(analysis.pos_only_params) != 1:
        raise TypeError(
            f"@node '{func.__name__}' requires exactly 1 positional-only argument (ctx), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return NodeFactory(func, analysis.specs, analysis.public_signature)


@overload
def node(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[NodeFunc[_P]], NodeFactory[_P]]: ...
@overload
def node(
    func: NodeFunc[_P], /, *, resolve_type_hints: bool = True
) -> NodeFactory[_P]: ...
def node(
    func: Union[NodeFunc[_P], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[NodeFunc[_P]], NodeFactory[_P]],
    NodeFactory[_P],
]:
    if func is _MISSING:
        return partial(_node, resolve_type_hints=resolve_type_hints)
    else:
        return _node(func, resolve_type_hints=resolve_type_hints)


def _expression(
    func: ExpressionFunc[_P, _R], /, *, resolve_type_hints: bool
) -> ExpressionFactory[_P, _R]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    if len(analysis.pos_only_params) != 1:
        raise TypeError(
            f"@expression '{func.__name__}' requires exactly 1 positional-only argument (ctx), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return ExpressionFactory(func, analysis.specs, analysis.public_signature)


@overload
def expression(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[ExpressionFunc[_P, _R]], ExpressionFactory[_P, _R]]: ...
@overload
def expression(
    func: ExpressionFunc[_P, _R], /, *, resolve_type_hints: bool = True
) -> ExpressionFactory[_P, _R]: ...
def expression(
    func: Union[ExpressionFunc[_P, _R], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[ExpressionFunc[_P, _R]], ExpressionFactory[_P, _R]],
    ExpressionFactory[_P, _R],
]:
    if func is _MISSING:
        return partial(_expression, resolve_type_hints=resolve_type_hints)
    else:
        return _expression(func, resolve_type_hints=resolve_type_hints)


def _wrapper(
    func: WrapperFunc[_P], /, *, resolve_type_hints: bool
) -> WrapperFactory[_P]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    # Wrapper signature: (ctx, wrapped, call_next, **kwargs)
    # The first 3 arguments should be positional-only.
    if len(analysis.pos_only_params) != 3:
        raise TypeError(
            f"@wrapper '{func.__name__}' requires exactly 3 positional-only arguments (ctx, wrapped, call_next), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return WrapperFactory(func, analysis.specs, analysis.public_signature)


@overload
def wrapper(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[WrapperFunc[_P]], WrapperFactory[_P]]: ...
@overload
def wrapper(
    func: WrapperFunc[_P], /, *, resolve_type_hints: bool = True
) -> WrapperFactory[_P]: ...
def wrapper(
    func: Union[WrapperFunc[_P], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[WrapperFunc[_P]], WrapperFactory[_P]],
    WrapperFactory[_P],
]:
    if func is _MISSING:
        return partial(_wrapper, resolve_type_hints=resolve_type_hints)
    else:
        return _wrapper(func, resolve_type_hints=resolve_type_hints)


def _async_node(
    func: AsyncNodeFunc[_P], /, *, resolve_type_hints: bool
) -> AsyncNodeFactory[_P]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    if len(analysis.pos_only_params) != 1:
        raise TypeError(
            f"@async_node '{func.__name__}' requires exactly 1 positional-only argument (ctx), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return AsyncNodeFactory(func, analysis.specs, analysis.public_signature)


@overload
def async_node(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[AsyncNodeFunc[_P]], AsyncNodeFactory[_P]]: ...
@overload
def async_node(
    func: AsyncNodeFunc[_P], /, *, resolve_type_hints: bool = True
) -> AsyncNodeFactory[_P]: ...
def async_node(
    func: Union[AsyncNodeFunc[_P], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[AsyncNodeFunc[_P]], AsyncNodeFactory[_P]],
    AsyncNodeFactory[_P],
]:
    if func is _MISSING:
        return partial(_async_node, resolve_type_hints=resolve_type_hints)
    else:
        return _async_node(func, resolve_type_hints=resolve_type_hints)


def _async_expression(
    func: AsyncExpressionFunc[_P, _R], /, *, resolve_type_hints: bool
) -> AsyncExpressionFactory[_P, _R]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    if len(analysis.pos_only_params) != 1:
        raise TypeError(
            f"@async_expression '{func.__name__}' requires exactly 1 positional-only argument (ctx), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return AsyncExpressionFactory(func, analysis.specs, analysis.public_signature)


@overload
def async_expression(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[AsyncExpressionFunc[_P, _R]], AsyncExpressionFactory[_P, _R]]: ...
@overload
def async_expression(
    func: AsyncExpressionFunc[_P, _R], /, *, resolve_type_hints: bool = True
) -> AsyncExpressionFactory[_P, _R]: ...
def async_expression(
    func: Union[AsyncExpressionFunc[_P, _R], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[AsyncExpressionFunc[_P, _R]], AsyncExpressionFactory[_P, _R]],
    AsyncExpressionFactory[_P, _R],
]:
    if func is _MISSING:
        return partial(_async_expression, resolve_type_hints=resolve_type_hints)
    else:
        return _async_expression(func, resolve_type_hints=resolve_type_hints)


def _async_wrapper(
    func: AsyncWrapperFunc[_P], /, *, resolve_type_hints: bool
) -> AsyncWrapperFactory[_P]:
    analysis = analyze_signature(func, resolve_type_hints=resolve_type_hints)
    if len(analysis.pos_only_params) != 3:
        raise TypeError(
            f"@async_wrapper '{func.__name__}' requires exactly 3 positional-only arguments (ctx, wrapped, call_next), "
            f"but found {len(analysis.pos_only_params)}."
        )
    return AsyncWrapperFactory(func, analysis.specs, analysis.public_signature)


@overload
def async_wrapper(
    func: _Missing = _MISSING, /, *, resolve_type_hints: bool = True
) -> Callable[[AsyncWrapperFunc[_P]], AsyncWrapperFactory[_P]]: ...
@overload
def async_wrapper(
    func: AsyncWrapperFunc[_P], /, *, resolve_type_hints: bool = True
) -> AsyncWrapperFactory[_P]: ...
def async_wrapper(
    func: Union[AsyncWrapperFunc[_P], _Missing] = _MISSING,
    /,
    *,
    resolve_type_hints: bool = True,
) -> Union[
    Callable[[AsyncWrapperFunc[_P]], AsyncWrapperFactory[_P]],
    AsyncWrapperFactory[_P],
]:
    if func is _MISSING:
        return partial(_async_wrapper, resolve_type_hints=resolve_type_hints)
    else:
        return _async_wrapper(func, resolve_type_hints=resolve_type_hints)


from .tree import NODE_ENGINE, NODE_PREPARE_ENGINE
from .render import get_render_string
from .eval import prepare_eval_plan, execute_eval_plan, async_execute_eval_plan



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
slyme pytree utility module.
"""

from .core import (
    # Types & Keys
    KeyPath,
    PyTreeKey,
    SequenceKey,
    MappingKey,
    AttributeKey,
    CallKey,
    P,
    PyTreeAux,
    # Structure & Engine
    PyTreeDef,
    LeafDef,
    ContainerDef,
    TraverseAux,
    PyTreeEngine,
    # Instances
    PYTREE_ENGINE_REGISTRY,
)

__all__ = [
    "KeyPath",
    "PyTreeKey",
    "SequenceKey",
    "MappingKey",
    "AttributeKey",
    "CallKey",
    "P",
    "PyTreeAux",
    "PyTreeDef",
    "LeafDef",
    "ContainerDef",
    "TraverseAux",
    "PyTreeEngine",
    "PYTREE_ENGINE_REGISTRY",
]



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class NodeException(Exception):
    """Base exception class for all exceptions of ``Node``."""

    pass


class NodeTerminate(NodeException):
    """Terminate the whole node execution."""

    def __init__(self, msg: str = "Terminate.", source_node=None) -> None:
        super().__init__(msg, source_node)

    @property
    def msg(self):
        return self.args[0]

    @msg.setter
    def msg(self, value):
        self.args = (value, self.args[1])

    @property
    def source_node(self):
        return self.args[1]

    @source_node.setter
    def source_node(self, value):
        self.args = (self.args[0], value)

    def __str__(self) -> str:
        return f"source_node: {self.source_node}, msg: {self.msg}"


# Node exception records.
class NodeExceptionRecord(NodeException):
    """Used to record the node exception info."""

    def __init__(self, exception_node, exception: Exception) -> None:
        super().__init__(exception_node, exception)

    @property
    def exception_node(self):
        return self.args[0]

    @property
    def exception(self):
        return self.args[1]

    def __str__(self) -> str:
        return f"exception_node: {self.exception_node}"


class WrapperExceptionRecord(NodeException):
    """Used to record the exception info raised by a ``Wrapper``."""

    def __init__(self, exception_node, wrapped_node, exception: Exception) -> None:
        super().__init__(exception_node, wrapped_node, exception)

    @property
    def exception_node(self):
        return self.args[0]

    @property
    def wrapped_node(self):
        return self.args[1]

    @property
    def exception(self):
        return self.args[2]

    def __str__(self) -> str:
        return f"exception_wrapper: {self.exception_node}, wrapped_node: {self.wrapped_node}"


class ExpressionExceptionRecord(NodeException):
    """Used to record the exception info raised by a ``Expression``."""

    def __init__(self, exception_node, exception: Exception, source_node=None) -> None:
        super().__init__(exception_node, exception, source_node)

    @property
    def exception_node(self):
        return self.args[0]

    @property
    def exception(self):
        return self.args[1]

    @property
    def source_node(self):
        return self.args[2]

    @source_node.setter
    def source_node(self, value):
        self.args = (self.args[0], self.args[1], value)

    def __str__(self) -> str:
        return f"exception_expression: {self.exception_node}, source_node: {self.source_node}"



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PyTree engine configuration and logic for Node.
"""

import types
from typing import Any, Iterable, cast
from types import MappingProxyType
from slyme.utils.pytree import (
    PyTreeEngine,
    PYTREE_ENGINE_REGISTRY,
    PyTreeAux,
    AttributeKey,
    MappingKey,
)
from slyme.utils.pytree.common import flatten_mapping_proxy, unflatten_mapping_proxy
from .core import (
    NodeDef,
    ExpressionDef,
    WrapperDef,
    NodeExec,
    ExpressionExec,
    WrapperExec,
    AsyncNodeDef,
    AsyncExpressionDef,
    AsyncWrapperDef,
    AsyncNodeExec,
    AsyncExpressionExec,
    AsyncWrapperExec,
)

# 1. Standard Engine: Preserves Types (Def -> Def, List -> List)
# Used for inspection, rendering, and validation.
NODE_ENGINE = PyTreeEngine("node_engine")
PYTREE_ENGINE_REGISTRY.register(NODE_ENGINE, key="node_engine")
# 2. Prepare Engine: Transforms Types (Def -> Exec, List -> Tuple, Dict -> MappingProxy)
# Used for compiling the definition tree into an execution tree.
NODE_PREPARE_ENGINE = PyTreeEngine("node_prepare", register_defaults=False)
PYTREE_ENGINE_REGISTRY.register(NODE_PREPARE_ENGINE, key="node_prepare")


# NodeDef Logic
def _flatten_node_def(obj: NodeDef) -> tuple[Iterable[Any], PyTreeAux]:
    """
    Flatten NodeDef into children (wrappers + kwargs values) and metadata.
    """
    # 1. Wrappers (List[WrapperDef]) treated as a single child container
    children = [obj.wrappers]
    rich_keys = [AttributeKey("wrappers")]

    # 2. Kwargs
    for k, v in obj._kwargs.items():
        children.append(v)
        rich_keys.append(MappingKey(k))

    metadata = {"func": obj._func, "specs": obj._specs}
    return tuple(children), PyTreeAux(
        children_keys=tuple(rich_keys), metadata=metadata, cls=NodeDef
    )


def _unflatten_node_def(children: Iterable[Any], aux: PyTreeAux) -> NodeDef:
    """
    Reconstruct NodeDef (Standard Engine).
    """
    # Use zip for cleaner iteration
    iterator = zip(aux.children_keys, children)
    # 1. Wrappers (Always the first element based on flatten logic)
    _, wrappers = next(iterator)
    # 2. Kwargs (Remaining elements)
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return NodeDef(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


def _unflatten_node_def_to_exec(children: Iterable[Any], aux: PyTreeAux) -> NodeExec:
    """
    Transform NodeDef into NodeExec (Prepare Engine).
    """
    iterator = zip(aux.children_keys, children)
    # 1. Wrappers
    # NOTE: The engine has already recursively transformed the wrappers list into a tuple.
    _, wrappers = next(iterator)
    # 2. Kwargs
    # NOTE: NodeExec.__init__ is responsible for converting this dict to MappingProxy.
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return NodeExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


# ExpressionDef Logic
def _flatten_expression_def(
    obj: ExpressionDef,
) -> tuple[Iterable[Any], PyTreeAux]:
    """
    Flatten ExpressionDef.
    """
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=ExpressionDef
    )


def _unflatten_expression_def(children: Iterable[Any], aux: PyTreeAux) -> ExpressionDef:
    """
    Reconstruct ExpressionDef (Standard Engine).
    """
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return ExpressionDef(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _unflatten_expression_def_to_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> ExpressionExec:
    """
    Transform ExpressionDef into ExpressionExec (Prepare Engine).
    """
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return ExpressionExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        kwargs=kwargs,
    )


# WrapperDef Logic
def _flatten_wrapper_def(obj: WrapperDef) -> tuple[Iterable[Any], PyTreeAux]:
    """
    Flatten WrapperDef.
    """
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=WrapperDef
    )


def _unflatten_wrapper_def(children: Iterable[Any], aux: PyTreeAux) -> WrapperDef:
    """
    Reconstruct WrapperDef (Standard Engine).
    """
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return WrapperDef(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _unflatten_wrapper_def_to_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> WrapperExec:
    """
    Transform WrapperDef into WrapperExec (Prepare Engine).
    """
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return WrapperExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        kwargs=kwargs,
    )


# Exec Types Logic
def _flatten_node_exec(obj: NodeExec) -> tuple[Iterable[Any], PyTreeAux]:
    children = [obj.wrappers]
    rich_keys = [AttributeKey("wrappers")]
    for k, v in obj._kwargs.items():
        children.append(v)
        rich_keys.append(MappingKey(k))
    metadata = {"func": obj._func, "specs": obj._specs}
    return tuple(children), PyTreeAux(
        children_keys=tuple(rich_keys), metadata=metadata, cls=NodeExec
    )


def _unflatten_node_exec(children: Iterable[Any], aux: PyTreeAux) -> NodeExec:
    iterator = zip(aux.children_keys, children)
    _, wrappers = next(iterator)
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return NodeExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


def _flatten_expression_exec(
    obj: ExpressionExec,
) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=ExpressionExec
    )


def _unflatten_expression_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> ExpressionExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return ExpressionExec(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _flatten_wrapper_exec(obj: WrapperExec) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=WrapperExec
    )


def _unflatten_wrapper_exec(children: Iterable[Any], aux: PyTreeAux) -> WrapperExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return WrapperExec(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


# AsyncNodeDef Logic
def _flatten_async_node_def(obj: AsyncNodeDef) -> tuple[Iterable[Any], PyTreeAux]:
    children = [obj.wrappers]
    rich_keys = [AttributeKey("wrappers")]
    for k, v in obj._kwargs.items():
        children.append(v)
        rich_keys.append(MappingKey(k))
    metadata = {"func": obj._func, "specs": obj._specs}
    return tuple(children), PyTreeAux(
        children_keys=tuple(rich_keys), metadata=metadata, cls=AsyncNodeDef
    )


def _unflatten_async_node_def(children: Iterable[Any], aux: PyTreeAux) -> AsyncNodeDef:
    iterator = zip(aux.children_keys, children)
    _, wrappers = next(iterator)
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return AsyncNodeDef(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


def _unflatten_async_node_def_to_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncNodeExec:
    iterator = zip(aux.children_keys, children)
    _, wrappers = next(iterator)
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return AsyncNodeExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


# AsyncExpressionDef Logic
def _flatten_async_expression_def(
    obj: AsyncExpressionDef,
) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=AsyncExpressionDef
    )


def _unflatten_async_expression_def(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncExpressionDef:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncExpressionDef(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _unflatten_async_expression_def_to_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncExpressionExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncExpressionExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        kwargs=kwargs,
    )


# AsyncWrapperDef Logic
def _flatten_async_wrapper_def(obj: AsyncWrapperDef) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=AsyncWrapperDef
    )


def _unflatten_async_wrapper_def(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncWrapperDef:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncWrapperDef(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _unflatten_async_wrapper_def_to_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncWrapperExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncWrapperExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        kwargs=kwargs,
    )


# AsyncExec Types Logic
def _flatten_async_node_exec(obj: AsyncNodeExec) -> tuple[Iterable[Any], PyTreeAux]:
    children = [obj.wrappers]
    rich_keys = [AttributeKey("wrappers")]
    for k, v in obj._kwargs.items():
        children.append(v)
        rich_keys.append(MappingKey(k))
    metadata = {"func": obj._func, "specs": obj._specs}
    return tuple(children), PyTreeAux(
        children_keys=tuple(rich_keys), metadata=metadata, cls=AsyncNodeExec
    )


def _unflatten_async_node_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncNodeExec:
    iterator = zip(aux.children_keys, children)
    _, wrappers = next(iterator)
    kwargs = {cast("MappingKey", k).key: v for k, v in iterator}
    return AsyncNodeExec(
        func=aux.metadata["func"],
        specs=aux.metadata["specs"],
        wrappers=wrappers,
        kwargs=kwargs,
    )


def _flatten_async_expression_exec(
    obj: AsyncExpressionExec,
) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=AsyncExpressionExec
    )


def _unflatten_async_expression_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncExpressionExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncExpressionExec(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


def _flatten_async_wrapper_exec(
    obj: AsyncWrapperExec,
) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(obj._kwargs.keys())
    children = tuple(obj._kwargs.values())
    rich_keys = tuple(MappingKey(k) for k in keys)
    metadata = {"func": obj._func, "specs": obj._specs}
    return children, PyTreeAux(
        children_keys=rich_keys, metadata=metadata, cls=AsyncWrapperExec
    )


def _unflatten_async_wrapper_exec(
    children: Iterable[Any], aux: PyTreeAux
) -> AsyncWrapperExec:
    kwargs = {cast("MappingKey", k).key: v for k, v in zip(aux.children_keys, children)}
    return AsyncWrapperExec(
        func=aux.metadata["func"], specs=aux.metadata["specs"], kwargs=kwargs
    )


# Container Logic (Prepare Transformations)
def _flatten_list(l: list) -> tuple[Iterable[Any], PyTreeAux]:
    return iter(l), PyTreeAux()


def _unflatten_to_tuple(children: Iterable[Any], _: PyTreeAux) -> tuple:
    return tuple(children)


def _flatten_tuple(t: tuple) -> tuple[Iterable[Any], PyTreeAux]:
    return iter(t), PyTreeAux()


def _unflatten_tuple(children: Iterable[Any], _: PyTreeAux) -> tuple:
    return tuple(children)


def _flatten_dict(d: dict) -> tuple[Iterable[Any], PyTreeAux]:
    keys = tuple(d.keys())
    rich_keys = tuple(MappingKey(k) for k in keys)
    children = (d[k] for k in keys)
    return children, PyTreeAux(children_keys=rich_keys)


def _unflatten_to_mapping_proxy(
    children: Iterable[Any], aux: PyTreeAux
) -> types.MappingProxyType:
    raw_keys = [k.key for k in cast("Iterable[MappingKey]", aux.children_keys)]
    return types.MappingProxyType(dict(zip(raw_keys, children)))


# Registrations
# --- NODE_PYTREE_ENGINE (Def -> Def, Exec -> Exec) ---
NODE_ENGINE.register(NodeDef, _flatten_node_def, _unflatten_node_def, strict=True)
NODE_ENGINE.register(
    ExpressionDef, _flatten_expression_def, _unflatten_expression_def, strict=True
)
NODE_ENGINE.register(
    WrapperDef, _flatten_wrapper_def, _unflatten_wrapper_def, strict=True
)
NODE_ENGINE.register(NodeExec, _flatten_node_exec, _unflatten_node_exec, strict=True)
NODE_ENGINE.register(
    ExpressionExec,
    _flatten_expression_exec,
    _unflatten_expression_exec,
    strict=True,
)
NODE_ENGINE.register(
    WrapperExec, _flatten_wrapper_exec, _unflatten_wrapper_exec, strict=True
)
NODE_ENGINE.register(
    AsyncNodeDef, _flatten_async_node_def, _unflatten_async_node_def, strict=True
)
NODE_ENGINE.register(
    AsyncExpressionDef,
    _flatten_async_expression_def,
    _unflatten_async_expression_def,
    strict=True,
)
NODE_ENGINE.register(
    AsyncWrapperDef,
    _flatten_async_wrapper_def,
    _unflatten_async_wrapper_def,
    strict=True,
)
NODE_ENGINE.register(
    AsyncNodeExec, _flatten_async_node_exec, _unflatten_async_node_exec, strict=True
)
NODE_ENGINE.register(
    AsyncExpressionExec,
    _flatten_async_expression_exec,
    _unflatten_async_expression_exec,
    strict=True,
)
NODE_ENGINE.register(
    AsyncWrapperExec,
    _flatten_async_wrapper_exec,
    _unflatten_async_wrapper_exec,
    strict=True,
)
NODE_ENGINE.register(MappingProxyType, flatten_mapping_proxy, unflatten_mapping_proxy)

# --- NODE_PREPARE_PYTREE_ENGINE (Def -> Exec, Mutables -> Immutables) ---
# Custom Containers
NODE_PREPARE_ENGINE.register(list, _flatten_list, _unflatten_to_tuple)
NODE_PREPARE_ENGINE.register(tuple, _flatten_tuple, _unflatten_tuple)
NODE_PREPARE_ENGINE.register(dict, _flatten_dict, _unflatten_to_mapping_proxy)
# Def -> Exec Transformations
NODE_PREPARE_ENGINE.register(
    NodeDef, _flatten_node_def, _unflatten_node_def_to_exec, strict=True
)
NODE_PREPARE_ENGINE.register(
    ExpressionDef,
    _flatten_expression_def,
    _unflatten_expression_def_to_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    WrapperDef,
    _flatten_wrapper_def,
    _unflatten_wrapper_def_to_exec,
    strict=True,
)
# Exec Identity (Exec -> Exec)
NODE_PREPARE_ENGINE.register(
    NodeExec, _flatten_node_exec, _unflatten_node_exec, strict=True
)
NODE_PREPARE_ENGINE.register(
    ExpressionExec,
    _flatten_expression_exec,
    _unflatten_expression_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    WrapperExec, _flatten_wrapper_exec, _unflatten_wrapper_exec, strict=True
)
NODE_PREPARE_ENGINE.register(
    AsyncNodeDef,
    _flatten_async_node_def,
    _unflatten_async_node_def_to_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    AsyncExpressionDef,
    _flatten_async_expression_def,
    _unflatten_async_expression_def_to_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    AsyncWrapperDef,
    _flatten_async_wrapper_def,
    _unflatten_async_wrapper_def_to_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    AsyncNodeExec, _flatten_async_node_exec, _unflatten_async_node_exec, strict=True
)
NODE_PREPARE_ENGINE.register(
    AsyncExpressionExec,
    _flatten_async_expression_exec,
    _unflatten_async_expression_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    AsyncWrapperExec,
    _flatten_async_wrapper_exec,
    _unflatten_async_wrapper_exec,
    strict=True,
)
NODE_PREPARE_ENGINE.register(
    MappingProxyType, flatten_mapping_proxy, unflatten_mapping_proxy
)



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import inspect
import types
from enum import Enum
from dataclasses import dataclass
from typing import (
    Callable,
    Any,
    Union,
    Annotated,
    get_type_hints,
    get_origin,
    get_args,
    Mapping,
    Sequence,
    TypeVar,
)
from collections import ChainMap
from slyme.utils.exception import enrich_exception
from slyme.context import Context

__all__ = [
    "spec",
    "Auto",
]
T = TypeVar("T")
_Missing = Enum("_Missing", ["MARK"])
_MISSING = _Missing.MARK
EvaluatorFunc = Callable[[Context], Any]


# Spec Definitions
@dataclass(frozen=True)
class Spec:
    """
    Dependency injection metadata for functional node parameters.
    """

    default: Union[Any, _Missing] = _MISSING
    default_factory: Union[Callable[[], Any], _Missing] = _MISSING
    auto_eval: Union[bool, _Missing] = _MISSING

    def __post_init__(self):
        if self.default is not _MISSING and self.default_factory is not _MISSING:
            raise ValueError(
                "`default` and `default_factory` cannot be set at the same time."
            )

    def _build(self, value: Any = _MISSING) -> Any:
        if value is not _MISSING:
            return value
        if self.default is not _MISSING:
            return self.default
        if self.default_factory is not _MISSING:
            return self.default_factory()
        raise ValueError("Missing required parameter.")

    def should_eval(self, value: Any) -> bool:
        if self.auto_eval is _MISSING:
            raise ValueError(
                "`auto_eval` should not be `_MISSING` when `should_eval` is called."
            )
        if not self.auto_eval:
            return False
        from slyme.node.eval import contains_eval_type

        return contains_eval_type(value)


def spec(
    default: Union[Any, _Missing] = _MISSING,
    default_factory: Union[Callable[[], Any], _Missing] = _MISSING,
    auto_eval: Union[bool, _Missing] = _MISSING,
) -> Any:
    return Spec(default=default, default_factory=default_factory, auto_eval=auto_eval)


# Some syntactic sugar
Auto = Annotated[T, Spec(auto_eval=True)]


# Inspection & Signature Analysis
@dataclass(frozen=True)
class SignatureAnalysis:
    pos_only_params: tuple[inspect.Parameter, ...]
    public_signature: inspect.Signature
    specs: Mapping[str, Spec]

    def __post_init__(self):
        if not isinstance(self.specs, types.MappingProxyType):
            object.__setattr__(self, "specs", types.MappingProxyType(self.specs))


def _collect_specs_from_hint(hint: Any) -> list[Spec]:
    """Recursively collect Spec annotations from a type hint."""
    # NOTE: Compatibility for Python < 3.11: get_type_hints auto-wraps parameters
    # with None defaults in Optional. Safely unwrap this outer Optional/Union
    # to expose the underlying Annotated type.
    if sys.version_info < (3, 11):
        if get_origin(hint) is Union:
            args = get_args(hint)
            if len(args) == 2 and type(None) in args:
                hint = args[0] if args[1] is type(None) else args[1]

    specs = []
    current = hint

    while get_origin(current) is Annotated:
        args = get_args(current)
        for item in args[1:]:
            if isinstance(item, Spec):
                specs.append(item)
        current = args[0]

    return specs


def resolve_spec(param: inspect.Parameter, hint: Any) -> Spec:
    if hint is None:
        collected_specs = []
    else:
        collected_specs = _collect_specs_from_hint(hint)

    if param.default is not inspect.Parameter.empty:
        if isinstance(param.default, Spec):
            collected_specs.append(param.default)
        else:
            collected_specs.append(Spec(default=param.default))

    merged_values: dict[str, Any] = {
        "default": _MISSING,
        "default_factory": _MISSING,
        "auto_eval": _MISSING,
    }

    for s in collected_specs:
        for field in merged_values:
            if (val := getattr(s, field)) is _MISSING:
                continue
            if merged_values[field] is not _MISSING:
                raise ValueError(
                    f"Conflict: Multiple definitions for '{field}' in parameter '{param.name}'."
                )
            merged_values[field] = val

    if merged_values["auto_eval"] is _MISSING:
        merged_values["auto_eval"] = False

    return Spec(**merged_values)


def analyze_signature(
    func: Callable, *, resolve_type_hints: bool = True
) -> SignatureAnalysis:
    """
    Analyze the function signature to separate runtime parameters and config parameters.
    """
    sig = inspect.signature(func)
    params = list(sig.parameters.values())

    pos_only_params = []
    public_params = []  # Used for factory signature
    specs: dict[str, Spec] = {}

    if resolve_type_hints:
        type_hints = get_type_hints(func, include_extras=True)
    else:
        type_hints = {}

    for p in params:
        with enrich_exception(f"in definition of '{func.__name__}'"):
            if p.kind == inspect.Parameter.POSITIONAL_ONLY:
                pos_only_params.append(p)
            elif p.kind == inspect.Parameter.KEYWORD_ONLY:
                public_params.append(p)
                # Spec Resolution Logic: Always returns a Spec object now
                specs[p.name] = resolve_spec(p, type_hints.get(p.name))
            else:
                kind_name = str(p.kind)
                raise TypeError(
                    f"Invalid parameter '{p.name}' of kind {kind_name}. "
                    f"Functional nodes strict rules:\n"
                    f"  1. Runtime args (e.g. ctx) must be POSITIONAL_ONLY (before '/').\n"
                    f"  2. Config args must be KEYWORD_ONLY (after '*')."
                )

    public_signature = sig.replace(parameters=public_params)
    return SignatureAnalysis(
        pos_only_params=tuple(pos_only_params),
        public_signature=public_signature,
        specs=types.MappingProxyType(specs),
    )


def process_kwargs(
    specs: Mapping[str, Spec],
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    """
    Validate and process kwargs:
    1. Check for unexpected arguments.
    2. Check for missing required arguments.
    3. Inject Spec-resolved values.
    """
    allowed_names = set(specs.keys())
    input_names = set(kwargs.keys())

    # 1. Strict Subset Check
    unknown_args = input_names - allowed_names
    if unknown_args:
        raise TypeError(
            f"Got unexpected keyword argument(s) {list(unknown_args)}. "
            f"Allowed arguments: {list(allowed_names)}."
        )

    # 2. Apply Specs
    final_kwargs = {}
    for name, spec_obj in specs.items():
        value = kwargs.get(name, _MISSING)
        with enrich_exception(f"for parameter '{name}'"):
            value = spec_obj._build(value)
        final_kwargs[name] = value

    return final_kwargs


def resolve_arguments(
    specs: Mapping[str, Spec],
    scopes: Sequence[Mapping[str, Any]],
    overrides: Mapping[str, Any],
) -> dict[str, Any]:
    """
    Resolve arguments from scopes and overrides based on specs.
    """
    # 1. Create a unified lookup map: overrides > last scope > ... > first scope
    # ChainMap looks up keys in the first mapping, then the second, and so on.
    unified_map = ChainMap(overrides, *reversed(scopes))
    # 2. Iterate through required parameters defined in the specs
    return {name: unified_map[name] for name in specs if name in unified_map}



# Copyright 2026 Slymer-Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Union
from collections.abc import Iterable, Sequence
from slyme.context import Context
from .core import Node, node, AsyncNode, async_node

__all__ = [
    "sequential_exec",
    "sequential",
    "async_sequential_exec",
    "async_sequential",
]


def sequential_exec(ctx: Context, nodes: Iterable[Node]) -> Context:
    """
    Sequentially execute a list of nodes, passing the context from one to the next.

    Args:
        ctx (Context): The initial context to pass through the nodes.
        nodes (Iterable[Node]): An iterable of nodes to execute.

    Returns:
        Context: The final context after all nodes have been executed.
    """
    for node_ in nodes:
        ctx = node_(ctx)
    return ctx


@node
def sequential(ctx: Context, /, *, nodes: Sequence[Node]) -> Context:
    """
    Sequentially execute a list of nodes, passing the context from one to the next.

    Args:
        ctx (Context): The initial context to pass through the nodes.
        nodes (Sequence[Node]): A sequence of nodes to execute.

    Returns:
        Context: The final context after all nodes have been executed.
    """
    return sequential_exec(ctx, nodes)


async def async_sequential_exec(
    ctx: Context, nodes: Iterable[Union[Node, AsyncNode]]
) -> Context:
    """
    Sequentially execute a list of nodes, passing the context from one to the next.

    Args:
        ctx (Context): The initial context to pass through the nodes.
        nodes (Iterable[Union[Node, AsyncNode]]): An iterable of nodes to execute.

    Returns:
        Context: The final context after all nodes have been executed.
    """
    for node_ in nodes:
        if isinstance(node_, AsyncNode):
            ctx = await node_(ctx)
        else:
            ctx = await asyncio.to_thread(node_, ctx)
    return ctx


@async_node
async def async_sequential(
    ctx: Context, /, *, nodes: Sequence[Union[Node, AsyncNode]]
) -> Context:
    """
    Sequentially execute a list of nodes, passing the context from one to the next.

    Args:
        ctx (Context): The initial context to pass through the nodes.
        nodes (Sequence[Union[Node, AsyncNode]]): A sequence of nodes to execute.

    Returns:
        Context: The final context after all nodes have been executed.
    """
    return await async_sequential_exec(ctx, nodes)
